# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (C) 2010 Jonathan Matthew (replay gain code)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio, GObject

from gi.repository.Gio import FILE_ATTRIBUTE_FILESYSTEM_SIZE, \
                              FILE_ATTRIBUTE_FILESYSTEM_FREE

from gettext import gettext as _

from lollypop.logger import Logger
from lollypop.define import App, Type
from lollypop.sync_mtp import MtpSync
from lollypop.utils import emit_signal


class DeviceWidget(Gtk.ListBoxRow):
    """
        A device widget for sync
    """

    __gsignals__ = {
        "syncing": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self, name, uri, icon=None):
        """
            Init widget
            @param name as str
            @param uri as str
            @param icon as Gio.Icon
        """
        Gtk.ListBoxRow.__init__(self)
        self.get_style_context().add_class("background")
        self.connect("map", self.__on_map)
        self.__name = name
        self.__uri = uri
        self.__progress = 0
        self.__builder = Gtk.Builder()
        self.__builder.add_from_resource("/org/gnome/Lollypop/DeviceWidget.ui")
        self.__progressbar = self.__builder.get_object("progress")
        self.__revealer = self.__builder.get_object("revealer")
        self.__switch_normalize = self.__builder.get_object("switch_normalize")
        self.__builder.get_object("name").set_label(self.__name)
        self.__combobox = self.__builder.get_object("combobox")
        self.__sync_button = self.__builder.get_object("sync_button")
        if icon is not None:
            device_symbolic = self.__builder.get_object("device-symbolic")
            device_symbolic.set_from_gicon(icon, Gtk.IconSize.DND)
        self.add(self.__builder.get_object("widget"))
        self.__builder.connect_signals(self)
        self.__calculate_free_space()
        self.__mtp_sync = MtpSync()
        self.__mtp_sync.connect("sync-finished", self.__on_sync_finished)
        self.__mtp_sync.connect("sync-progress", self.__on_sync_progress)
        for encoder in self.__mtp_sync._GST_ENCODER.keys():
            if not self.__mtp_sync.check_encoder_status(encoder):
                self.__builder.get_object(encoder).set_sensitive(False)
        self.connect("destroy", self.__on_destroy)

    @property
    def uri(self):
        """
            Get device URI
            @return str
        """
        return self.__uri

    @property
    def name(self):
        """
            Get device name
            @return str
        """
        return self.__name

    @property
    def progress(self):
        """
            Get progress status
            @return int
        """
        return self.__progress

#######################
# PROTECTED           #
#######################
    def _on_reveal_button_clicked(self, button):
        """
            Show advanced device options
            @param button as Gtk.Button
        """
        revealed = self.__revealer.get_reveal_child()
        self.__revealer.set_reveal_child(not revealed)
        if revealed:
            button.get_image().get_style_context().remove_class("image-reveal")
        else:
            button.get_image().get_style_context().add_class("image-reveal")

    def _on_content_albums_clicked(self, button):
        """
            Show synced albums
            @param button as Gtk.Button
        """
        index = self.__get_device_index()
        if index is not None:
            App().window.container.show_view([Type.DEVICE_ALBUMS], index)
        else:
            App().notify.send("Lollypop", _("No synchronized albums"))

    def _on_content_playlists_clicked(self, button):
        """
            Show synced playlists
            @param button as Gtk.Button
        """
        index = self.__get_device_index()
        if index is not None:
            App().window.container.show_view([Type.DEVICE_PLAYLISTS], index)
        else:
            App().notify.send("Lollypop", _("No synchronized playlists"))

    def _on_sync_button_clicked(self, button):
        """
            Sync music on device
            @param button as Gtk.Button
        """
        if self.__sync_button.get_label() == _("Synchronize"):
            self.__progress = 0
            uri = self.__get_music_uri()
            index = self.__get_device_index()
            if index is not None:
                App().task_helper.run(self.__mtp_sync.sync, uri, index)
                emit_signal(self, "syncing", True)
                button.set_label(_("Cancel"))
        else:
            self.__mtp_sync.cancel()
            button.set_sensitive(False)

    def _on_convert_toggled(self, widget):
        """
            Save option
            @param widget as Gtk.RadioButton
        """
        if widget.get_active():
            encoder = widget.get_name()
            if encoder == "convert_none":
                self.__switch_normalize.set_sensitive(False)
                self.__mtp_sync.db.set_normalize(False)
                self.__mtp_sync.db.set_encoder("convert_none")
            else:
                self.__switch_normalize.set_sensitive(True)
                self.__mtp_sync.db.set_encoder(encoder)
        self.__mtp_sync.db.save()

    def _on_normalize_state_set(self, widget, state):
        """
            Save option
            @param widget as Gtk.Switch
            @param state as bool
        """
        self.__mtp_sync.db.set_normalize(state)
        self.__mtp_sync.db.save()

    def _on_combobox_changed(self, combobox):
        """
            Update DB
            @param combobox as Gtk.ComboxText
        """
        self.__load_uri_settings(self.__get_music_uri())

#######################
# PRIVATE             #
#######################
    def __get_device_index(self):
        """
            Get current device index
            @return int/None
        """
        try:
            devices = list(App().settings.get_value("devices"))
            index = devices.index(self.__name) + 1
        except Exception as e:
            Logger.warning("DeviceWidget::__get_device_index(): %s",
                           e)
            index = None
        return index

    def __get_music_uri(self):
        """
            Get music URI on device
            @return str
        """
        if self.__combobox.get_visible():
            if self.__uri.endswith("/"):
                uri = "%s%s/Music" % (self.__uri,
                                      self.__combobox.get_active_text())
            else:
                uri = "%s/%s/Music" % (self.__uri,
                                       self.__combobox.get_active_text())
        else:
            uri = "%s/Music" % self.__uri
        return uri

    def __get_basename_for_sync(self):
        """
            Get basename base on device content
            @return str
        """
        names = []
        try:
            if not self.__uri.startswith("mtp://") and\
                    self.__name != "Librem phone":
                return None

            # Search for previous sync
            d = Gio.File.new_for_uri(self.__uri)
            infos = d.enumerate_children(
                "standard::name,standard::type",
                Gio.FileQueryInfoFlags.NONE,
                None)

            for info in infos:
                if info.get_file_type() != Gio.FileType.DIRECTORY:
                    continue
                f = infos.get_child(info)
                uri = f.get_uri() + "/Music"
                previous_sync = Gio.File.new_for_uri("%s/unsync" % uri)
                if previous_sync.query_exists():
                    names.insert(0, info.get_name())
                else:
                    names.append(info.get_name())
            infos.close(None)
        except Exception as e:
            Logger.error("DeviceWidget::__get_best_uri_for_sync: %s: %s"
                         % (self.__uri, e))
        return names

    def __set_combobox_content(self, names):
        """
            Set combobox content based on names
            @param names as [str]
        """
        if self.__combobox.get_active_text():
            return
        self.__sync_button.set_sensitive(True)
        if names is None:
            self.__load_uri_settings(self.__get_music_uri())
            self.__combobox.hide()
        elif names:
            self.__combobox.show()
            for name in names:
                self.__combobox.append_text(name)
            self.__combobox.set_active(0)
        else:
            self.__sync_button.set_sensitive(False)
            self.__combobox.hide()

    def __load_uri_settings(self, uri):
        """
            Load settings at URI
            @param uri as str
        """
        self.__mtp_sync.db.load(uri)
        encoder = self.__mtp_sync.db.encoder
        normalize = self.__mtp_sync.db.normalize
        self.__switch_normalize.set_sensitive(False)
        self.__switch_normalize.set_active(normalize)
        self.__builder.get_object(encoder).set_active(True)
        for encoder in self.__mtp_sync._GST_ENCODER.keys():
            if not self.__mtp_sync.check_encoder_status(encoder):
                self.__builder.get_object(encoder).set_sensitive(False)

    def __calculate_free_space(self):
        """
            Calculate free space on device
        """
        f = Gio.File.new_for_uri(self.__uri)
        f.query_filesystem_info_async("{},{}".format(
                                       FILE_ATTRIBUTE_FILESYSTEM_SIZE,
                                       FILE_ATTRIBUTE_FILESYSTEM_FREE),
                                      GLib.PRIORITY_DEFAULT,
                                      None,
                                      self.__on_filesystem_info)

    def __on_filesystem_info(self, source, result):
        """
            Show available space on disk
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        try:
            info = source.query_filesystem_info_finish(result)
            size = info.get_attribute_uint64(FILE_ATTRIBUTE_FILESYSTEM_SIZE)
            free = info.get_attribute_uint64(FILE_ATTRIBUTE_FILESYSTEM_FREE)
            if size == 0:
                return
            used = size - free
            fraction = 1 * used / size
            self.__progressbar.set_fraction(fraction)
            style_context = self.__progressbar.get_style_context()
            style_context.remove_class("usagebar-green")
            style_context.remove_class("usagebar-orange")
            style_context.remove_class("usagebar-red")
            if fraction < 0.6:
                style_context.add_class("usagebar-green")
            elif fraction < 0.8:
                style_context.add_class("usagebar-orange")
            else:
                style_context.add_class("usagebar-red")
        except Exception as e:
            Logger.error("DeviceWiget::__on_filesystem_info(): %s", e)

    def __on_sync_progress(self, mtp_sync, value):
        """
            Update progress bar
            @param mtp_sync as MtpSync
            @param value as float
        """
        self.__progress = value

    def __on_sync_finished(self, mtp_sync):
        """
            Emit finished signal
            @param mtp_sync as MtpSync
        """
        emit_signal(self, "syncing", False)
        self.__progress = 0
        self.__sync_button.set_label(_("Synchronize"))
        self.__sync_button.set_sensitive(True)
        self.__calculate_free_space()

    def __on_map(self, widget):
        """
            Setup combobox
            @param widget as Gtk.Widget
        """
        App().task_helper.run(self.__get_basename_for_sync,
                              callback=(self.__set_combobox_content,))

    def __on_destroy(self, widget):
        """
            Remove ref cycle
            @param widget as Gtk.Widget
        """
        self.__builder = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from lollypop.define import ViewType, MARGIN, App
from lollypop.utils import set_cursor_type
from lollypop.widgets_banner_artist import ArtistBannerWidget
from lollypop.widgets_album import AlbumWidget
from lollypop.objects_album import Album
from lollypop.view_lazyloading import LazyLoadingView
from lollypop.helper_size_allocation import SizeAllocationHelper
from lollypop.helper_signals import SignalsHelper, signals_map


class ArtistViewList(LazyLoadingView, SignalsHelper, SizeAllocationHelper):
    """
        Show artist albums in a list with tracks
    """

    @signals_map
    def __init__(self, genre_ids, artist_ids, storage_type, view_type):
        """
            Init ArtistView
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        LazyLoadingView.__init__(self, storage_type,
                                 view_type |
                                 ViewType.OVERLAY |
                                 ViewType.ARTIST)
        self.__boxes = []
        self.__others_boxes = []
        self.__width = 0
        self.__boxes_count = 0
        self.__current_box = 0
        self.__albums_count = 0
        self.__hovered_child = None
        self.__genre_ids = genre_ids
        self.__artist_ids = artist_ids
        self.__storage_type = storage_type
        self.__banner = ArtistBannerWidget(genre_ids, artist_ids,
                                           storage_type, self.view_type)
        self.__banner.show()
        self.__main_grid = Gtk.Grid.new()
        self.__main_grid.show()
        self.__main_grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__boxes_grid = Gtk.Grid.new()
        self.__boxes_grid.show()
        self.__boxes_grid.set_valign(Gtk.Align.START)
        self.__boxes_grid.set_halign(Gtk.Align.CENTER)
        self.__main_grid.add(self.__boxes_grid)
        self.add_widget(self.__main_grid, self.__banner)
        self.connect("populated", self.__on_populated)
        if App().settings.get_value("force-single-column"):
            self.__column_width = 1200
        else:
            self.__column_width = 600
        if App().animations:
            self.__event_controller = Gtk.EventControllerMotion.new(self)
            self.__event_controller.connect("motion", self.__on_motion)
        SizeAllocationHelper.__init__(self)
        return [
            (App().player, "current-changed", "_on_current_changed"),
        ]

    def populate(self):
        pass

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"genre_ids": self.__genre_ids,
                "artist_ids": self.__artist_ids,
                "storage_type": self.storage_type,
                "view_type": self.view_type}

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        filtered = []
        boxes_count = len(self.__boxes)
        for i in range(0, boxes_count):
            for child in self.__boxes[i].get_children():
                if isinstance(child, AlbumWidget):
                    filtered.append(child)
                    filtered += child.filtered
                else:
                    filtered += child.children
        return filtered

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return self.__banner.height + MARGIN

#######################
# PROTECTED           #
#######################
    def _get_child(self, album_id):
        """
            Get an album view widget
            @param album_id as int
            @return AlbumView
        """
        if self.destroyed:
            return None
        album = Album(album_id, self.__genre_ids, self.__artist_ids)
        widget = AlbumWidget(album,
                             self.storage_type,
                             ViewType.ARTIST)
        widget.show()
        widget.set_property("valign", Gtk.Align.START)
        box = self.__get_box_at_index(self.__current_box)
        box.add(widget)
        box.show()
        self.__current_box += 1
        if self.__current_box == self.__boxes_count:
            self.__current_box = 0
        return widget

    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
            @return bool
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            if allocation.width != self.__width:
                self.__width = allocation.width
                boxes_count = self.__width // self.__column_width
                if boxes_count < 1:
                    boxes_count = 1
                if self.__boxes_count == boxes_count:
                    return
                self.__boxes_count = boxes_count
                # Rework content
                if self.is_populated:
                    children = self.__get_children_sorted()
                    self.__remove_children()
                    self.__populate(children)
                else:
                    album_ids = App().albums.get_ids(self.__genre_ids,
                                                     self.__artist_ids,
                                                     self.storage_type,
                                                     True)
                    self.__albums_count = len(album_ids)
                    LazyLoadingView.populate(self, album_ids)
                return True
        return False

    def _on_current_changed(self, player):
        """
            Update children state
            @param player as Player
        """
        boxes_count = len(self.__boxes)
        for i in range(0, boxes_count):
            for child in self.__boxes[i].get_children():
                child.set_selection()

#######################
# PRIVATE             #
#######################
    def __get_box_at_index(self, index):
        """
            Get album box at index, add new ones if index out of range
            @param index as int
            @return Gtk.Box
        """
        while len(self.__boxes) <= index:
            box = Gtk.Box.new(Gtk.Orientation.VERTICAL, MARGIN)
            box.set_valign(Gtk.Align.START)
            box.set_property("margin", MARGIN)
            self.__boxes.append(box)
            self.__boxes_grid.add(box)
        return self.__boxes[index]

    def __populate(self, children):
        """
            Populate children
            @param children as [AlbumWidget]
        """
        self.__current_box = 0
        for child in children:
            box = self.__get_box_at_index(self.__current_box)
            box.show()
            box.add(child)
            self.__current_box += 1
            if self.__current_box == self.__boxes_count:
                self.__current_box = 0

    def __get_children_sorted(self):
        """
            Get children sorted (insert order)
            @return [Gtk.Widget]
        """
        boxes_count = len(self.__boxes)
        children = {}
        for i in range(0, boxes_count):
            children[i] = self.__boxes[i].get_children()
        sorted_children = []
        not_found = 0
        while not_found != boxes_count:
            not_found = 0
            for i in range(0, boxes_count):
                if children[i]:
                    child = children[i].pop(0)
                    sorted_children.append(child)
                else:
                    not_found += 1
        return sorted_children

    def __remove_children(self):
        """
            Remove children from boxes
        """
        boxes_count = len(self.__boxes)
        for i in range(0, boxes_count):
            self.__boxes[i].hide()
            children = self.__boxes[i].get_children()
            for child in children:
                self.__boxes[i].remove(child)

    def __unselect_selected(self):
        """
            Unselect selected child
        """
        if self.__hovered_child is not None:
            self.__hovered_child.unset_state_flags(
                Gtk.StateFlags.VISITED)
            set_cursor_type(self.__hovered_child.banner, "left_ptr")
            self.__hovered_child = None

    def __on_populated(self, view):
        """
            Add appears on albums
            @param view as ArtistViewBox
        """
        if self.__albums_count == 1:
            self.__boxes[0].get_children()[0].reveal_child()
        from lollypop.view_albums_line import AlbumsArtistAppearsOnLineView
        others_box = AlbumsArtistAppearsOnLineView(self.__artist_ids,
                                                   self.__genre_ids,
                                                   self.storage_type,
                                                   ViewType.SMALL |
                                                   ViewType.SCROLLED)
        others_box.set_margin_start(MARGIN)
        others_box.set_margin_end(MARGIN)
        others_box.populate()
        self.__main_grid.add(others_box)
        self.__others_boxes.append(others_box)

    def __on_motion(self, event_controller, x, y):
        """
            Update current selected child
            @param event_controller as Gtk.EventControllerMotion
            @param x as int
            @param y as int
        """
        hovered_child = None
        for i in range(0, len(self.__boxes)):
            for child in self.__boxes[i].get_children():
                (tx, ty) = child.translate_coordinates(self, 0, 0)
                width = child.get_allocated_width()
                height = child.get_allocated_height()
                if x > tx and x < tx + width and y > ty and y < ty + height:
                    hovered_child = child
                    break
        if hovered_child == self.__hovered_child:
            return
        elif hovered_child is not None:
            hovered_child.set_state_flags(Gtk.StateFlags.VISITED, False)
            self.__unselect_selected()
            self.__hovered_child = hovered_child
            set_cursor_type(hovered_child.banner)
        else:
            self.__unselect_selected()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

import sqlite3
from threading import Lock
from random import shuffle
import itertools

from lollypop.define import App, LOLLYPOP_DATA_PATH
from lollypop.database_upgrade import DatabaseAlbumsUpgrade
from lollypop.sqlcursor import SqlCursor
from lollypop.logger import Logger
from lollypop.localized import LocalizedCollation
from lollypop.utils import noaccents, sql_escape


class MyLock:
    """
        Lock with status
    """
    def __init__(self):
        self.__lock = Lock()
        self.__locked = False

    def acquire(self):
        self.__locked = True
        self.__lock.acquire()

    def release(self):
        self.__locked = False
        self.__lock.release()

    @property
    def locked(self):
        return self.__locked


class Database:
    """
        Base database object
    """

    DB_PATH = "%s/lollypop.db" % LOLLYPOP_DATA_PATH

    # SQLite documentation:
    # In SQLite, a column with type INTEGER PRIMARY KEY
    # is an alias for the ROWID.
    # Here, we define an id INT PRIMARY KEY but never feed it,
    # this make VACUUM not destroy rowids...
    __create_albums = """CREATE TABLE albums (id INTEGER PRIMARY KEY,
                                              name TEXT NOT NULL,
                                              mb_album_id TEXT,
                                              lp_album_id TEXT,
                                              no_album_artist BOOLEAN NOT NULL,
                                              year INT,
                                              timestamp INT,
                                              uri TEXT NOT NULL,
                                              popularity INT NOT NULL,
                                              rate INT NOT NULL,
                                              loved INT NOT NULL,
                                              mtime INT NOT NULL,
                                              storage_type INT NOT NULL,
                                              synced INT NOT NULL)"""
    __create_artists = """CREATE TABLE artists (id INTEGER PRIMARY KEY,
                                               name TEXT NOT NULL,
                                               sortname TEXT NOT NULL,
                                               mb_artist_id TEXT)"""
    __create_featuring = """CREATE TABLE featuring (
                                               artist_id INT NOT NULL,
                                               album_id INT NOT NULL)"""
    __create_genres = """CREATE TABLE genres (id INTEGER PRIMARY KEY,
                                            name TEXT NOT NULL)"""
    __create_album_artists = """CREATE TABLE album_artists (
                                                album_id INT NOT NULL,
                                                artist_id INT NOT NULL)"""
    __create_album_genres = """CREATE TABLE album_genres (
                                                album_id INT NOT NULL,
                                                genre_id INT NOT NULL)"""
    __create_album_timed_popularity = """CREATE TABLE albums_timed_popularity (
                                                album_id INT NOT NULL,
                                                mtime INT NOT NULL,
                                                popularity INT NOT NULL)"""
    __create_tracks = """CREATE TABLE tracks (id INTEGER PRIMARY KEY,
                                              name TEXT NOT NULL,
                                              uri TEXT NOT NULL,
                                              duration INT,
                                              tracknumber INT,
                                              discnumber INT,
                                              discname TEXT,
                                              album_id INT NOT NULL,
                                              year INT,
                                              timestamp INT,
                                              popularity INT NOT NULL,
                                              loved INT NOT NULL DEFAULT 0,
                                              rate INT NOT NULL,
                                              ltime INT NOT NULL,
                                              mtime INT NOT NULL,
                                              storage_type INT NOT NULL,
                                              mb_track_id TEXT,
                                              lp_track_id TEXT,
                                              bpm DOUBLE
                                              )"""
    __create_track_artists = """CREATE TABLE track_artists (
                                                track_id INT NOT NULL,
                                                artist_id INT NOT NULL)"""
    __create_track_genres = """CREATE TABLE track_genres (
                                                track_id INT NOT NULL,
                                                genre_id INT NOT NULL)"""
    __create_album_artists_idx = """CREATE index idx_aa ON album_artists(
                                                album_id)"""
    __create_track_artists_idx = """CREATE index idx_ta ON track_artists(
                                                track_id)"""
    __create_album_genres_idx = """CREATE index idx_ag ON album_genres(
                                                album_id)"""
    __create_track_genres_idx = """CREATE index idx_tg ON track_genres(
                                                track_id)"""

    def __init__(self):
        """
            Create database tables or manage update if needed
        """
        self.thread_lock = MyLock()
        f = Gio.File.new_for_path(self.DB_PATH)
        upgrade = DatabaseAlbumsUpgrade()
        if not f.query_exists():
            try:
                d = Gio.File.new_for_path(LOLLYPOP_DATA_PATH)
                if not d.query_exists():
                    d.make_directory_with_parents()
                # Create db schema
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_albums)
                    sql.execute(self.__create_artists)
                    sql.execute(self.__create_featuring)
                    sql.execute(self.__create_genres)
                    sql.execute(self.__create_album_genres)
                    sql.execute(self.__create_album_artists)
                    sql.execute(self.__create_album_timed_popularity)
                    sql.execute(self.__create_tracks)
                    sql.execute(self.__create_track_artists)
                    sql.execute(self.__create_track_genres)
                    sql.execute(self.__create_album_artists_idx)
                    sql.execute(self.__create_track_artists_idx)
                    sql.execute(self.__create_album_genres_idx)
                    sql.execute(self.__create_track_genres_idx)
                    sql.execute("PRAGMA user_version=%s" % upgrade.version)
            except Exception as e:
                Logger.error("Database::__init__(): %s" % e)
        else:
            upgrade.upgrade(self)

    def execute(self, request):
        """
            Execute SQL request (only smart one)
            @param request as str
            @return list
        """
        requests = []
        try:
            union_random = request.find("ORDER BY random()") != -1 and\
                request.find("UNION") != -1
            # Special case for UNION, does not support random()
            # Optimisation, split request and add LIMIT
            if union_random:
                request = request.replace("ORDER BY random()", "")
                limit_position = request.find("LIMIT")
                limit_str = request[limit_position:]
                limit_int = int(limit_str.replace("LIMIT ", ""))
                # Remove limit from main request
                request = request.replace(limit_str, "")
                request_split = request.split("UNION")
                sublimit = limit_int // len(request_split)
                for request in request_split:
                    request += " ORDER BY random() LIMIT %s" % sublimit
                    requests.append(request)
            else:
                requests = [request]
            ids = []
            with SqlCursor(App().db) as sql:
                for request in requests:
                    result = sql.execute(request)
                    ids += list(itertools.chain(*result))
                if union_random:
                    shuffle(ids)
                    return ids[:limit_int]
                else:
                    return ids
        except Exception as e:
            Logger.error("Database::execute(): %s -> %s", e, request)
        return []

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            c = sqlite3.connect(self.DB_PATH, 600.0)
            c.create_collation("LOCALIZED", LocalizedCollation())
            c.create_function("noaccents", 1, noaccents)
            c.create_function("sql_escape", 1, sql_escape)
            return c
        except:
            exit(-1)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib
from gi.repository.Gio import FILE_ATTRIBUTE_TIME_ACCESS

from time import time

from lollypop.logger import Logger
from lollypop.define import App, FileType


def get_file_type(uri):
    """
        Get file type from file extension
        @param uri as str
    """
    audio = ["3gp", "aa", "aac", "aax", "act", "aiff", "alac", "amr", "ape",
             "au", "awb", "dct", "dss", "dvf", "flac", "gsm", "iklax", "ivs",
             "m4a", "m4b", "m4p", "mmf", "mp3", "mpc", "msv", "nmf", "nsf",
             "ogg", "opus", "ra", "raw", "rf64", "sln", "tta", "voc", "vox",
             "wav", "wma", "wv", "webm", "8svx", "cda"]
    other = ["7z", "arj", "deb", "pkg", "rar", "rpm", "tar.gz", "z", "zip",
             "pdf", "doc", "docx", "xls", "xlsx", "ppt", "pptx", "db", "txt",
             "mov", "avi", "html", "ini", "cue", "nfo", "lrc"]
    image = ["ai", "bmp", "gif", "ico", "jpeg", "jpg",
             "png", "ps", "psd", "svg", "tif"]
    pls = ["pls", "m3u"]
    split = uri.lower().split(".")
    if len(split) < 2:
        return FileType.UNKNOWN
    elif split[-1] in audio:
        return FileType.AUDIO
    elif split[-1] in pls:
        return FileType.PLS
    elif split[-1] in image + other:
        return FileType.OTHER
    else:
        return FileType.UNKNOWN


def is_audio(info):
    """
        Return True if files is audio
        @param info as Gio.FileInfo
    """
    audio = ["application/ogg", "application/x-ogg", "application/x-ogm-audio",
             "audio/aac", "audio/mp4", "audio/mpeg", "audio/mpegurl",
             "audio/ogg", "audio/vnd.rn-realaudio", "audio/vorbis",
             "audio/x-flac", "audio/x-mp3", "audio/x-mpeg", "audio/x-mpegurl",
             "audio/x-ms-wma", "audio/x-musepack", "audio/x-oggflac",
             "audio/x-pn-realaudio", "application/x-flac", "audio/x-speex",
             "audio/x-vorbis", "audio/x-vorbis+ogg", "audio/x-wav",
             "x-content/audio-player", "audio/x-aac", "audio/m4a",
             "audio/x-m4a", "audio/mp3", "audio/ac3", "audio/flac",
             "audio/x-opus+ogg", "application/x-extension-mp4", "audio/x-ape",
             "audio/x-pn-aiff", "audio/x-pn-au", "audio/x-pn-wav",
             "audio/x-pn-windows-acm", "application/x-matroska",
             "audio/x-matroska", "audio/x-wavpack", "video/mp4",
             "audio/x-mod", "audio/x-mo3", "audio/x-xm", "audio/x-s3m",
             "audio/x-it", "audio/aiff", "audio/x-aiff"]
    if info is not None:
        if info.get_content_type() in audio:
            return True
    return False


def is_pls(info):
    """
        Return True if files is a playlist
        @param info as Gio.FileInfo
    """
    if info is not None:
        if info.get_content_type() in ["audio/x-mpegurl",
                                       "application/xspf+xml",
                                       "application/vnd.apple.mpegurl"]:
            return True
    return False


def get_mtime(info):
    """
        Return Last modified time of a given file
        @param info as Gio.FileInfo
    """
    mtime = info.get_attribute_as_string("time::modified")
    if mtime is None:
        return 0
    else:
        return int(mtime)


def remove_oldest(path, timestamp):
    """
        Remove oldest files at path
        @param path as str
        @param timestamp as int
    """
    SCAN_QUERY_INFO = "%s" % FILE_ATTRIBUTE_TIME_ACCESS
    try:
        d = Gio.File.new_for_path(path)
        infos = d.enumerate_children(SCAN_QUERY_INFO,
                                     Gio.FileQueryInfoFlags.NONE,
                                     None)
        for info in infos:
            f = infos.get_child(info)
            if info.get_file_type() == Gio.FileType.REGULAR:
                atime = int(info.get_attribute_as_string(
                    FILE_ATTRIBUTE_TIME_ACCESS))
                # File not used since one year
                if time() - atime > timestamp:
                    f.delete(None)
    except Exception as e:
        Logger.error("remove_oldest(): %s", e)


def is_readonly(uri):
    """
        Check if uri is readonly
    """
    try:
        if not uri:
            return True
        f = Gio.File.new_for_uri(uri)
        info = f.query_info("access::can-write",
                            Gio.FileQueryInfoFlags.NONE,
                            None)
        return not info.get_attribute_boolean("access::can-write")
    except:
        return True


def create_dir(path):
    """
        Create dir
        @param path as str
    """
    d = Gio.File.new_for_path(path)
    if not d.query_exists():
        try:
            d.make_directory_with_parents()
        except:
            Logger.info("Can't create %s" % path)


def install_youtube_dl():
    try:
        path = GLib.get_user_data_dir() + "/lollypop/python"
        argv = ["pip3", "install", "-t", path, "-U", "youtube-dl"]
        GLib.spawn_sync(None, argv, [], GLib.SpawnFlags.SEARCH_PATH, None)
    except Exception as e:
        Logger.error("install_youtube_dl: %s" % e)


def get_youtube_dl():
    """
        Get youtube-dl path and env
        @return (str, [])
    """
    if App().settings.get_value("recent-youtube-dl"):
        python_path = GLib.get_user_data_dir() + "/lollypop/python"
        path = "%s/bin/youtube-dl" % python_path
        env = ["PYTHONPATH=%s" % python_path]
        f = Gio.File.new_for_path(path)
        if f.query_exists():
            return (path, env)
    if GLib.find_program_in_path("youtube-dl"):
        return ("youtube-dl", [])
    else:
        return (None, [])


# From eyeD3 start
# eyeD3 is written and maintained by:
# Travis Shirk <travis@pobox.com>
def id3EncodingToString(encoding):
    from lollypop.define import LATIN1_ENCODING, UTF_8_ENCODING
    from lollypop.define import UTF_16_ENCODING, UTF_16BE_ENCODING
    if encoding == LATIN1_ENCODING:
        return "latin_1"
    elif encoding == UTF_8_ENCODING:
        return "utf_8"
    elif encoding == UTF_16_ENCODING:
        return "utf_16"
    elif encoding == UTF_16BE_ENCODING:
        return "utf_16_be"
    else:
        raise ValueError("Encoding unknown: %s" % encoding)


# From eyeD3 start
def decodeUnicode(bites, encoding):
    codec = id3EncodingToString(encoding)
    return bites.decode(codec)


def splitUnicode(data, encoding):
    from lollypop.define import UTF_16_ENCODING, UTF_16BE_ENCODING
    try:
        (d, t) = data.split(encoding, 1)
        # Try to fix invalid UTF16
        if encoding == UTF_16_ENCODING or encoding == UTF_16BE_ENCODING:
            if t[1] == 0:
                t = b''.join(t.split(b'\x00')) + b'\x00'
    except ValueError as e:
        Logger.warning("splitUnicode(): %s", e)
        (d, t) = data, b""
    return (d, t)
# From eyeD3 end



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from lollypop.assistant import Assistant


class GoogleAssistant(Assistant):
    """
        Google API assistant
    """

    def __init__(self):
        """
            Init assistant
        """
        rules = [
            {
              "title": _("What is Google API-key?"),
              "icon_name": "media-playback-start-symbolic",
              "markup": _("""
In order to use the YoutTube playback,
you need to create an API-Key for the

<b>"YouTube Data API v3"</b>

which is used to play songs."""),
              "uri_label": "",
              "uri": None,
              "right_button_label": _("Cancel"),
              "right_button_style": "",
              "left_button_label": _("Next"),
              "left_button_style": "suggested-action"
            },
            {
              "title": _("Connect to the console"),
              "icon_name": "web-browser-symbolic",
              "markup": _("""
<b>Open and log</b> in to the <b>Google Developer Console</b>"""),
              "uri_label": _("Access Google Developer Console"),
              "uri": "https://console.developers.google.com/apis/",
              "right_button_label": _("Back"),
              "right_button_style": "",
              "left_button_label": _("Next"),
              "left_button_style": "suggested-action"
            },
            {
              "title": _("Activate the API"),
              "icon_name": "folder-new-symbolic",
              "markup": _("""
<b>Add a new project</b> in the console.

After that, <b>add a new API</b> and search
for the <b>"YouTube Data API v3"</b>.

<b>Activate that API</b> for your project"""),
              "uri_label": "",
              "uri": None,
              "right_button_label": _("Back"),
              "right_button_style": "",
              "left_button_label": _("Next"),
              "left_button_style": "suggested-action"
            },
            {
              "title": _("Copy the key"),
              "icon_name": "dialog-password-symbolic",
              "markup": _("""
Open <b>"Credentials"</b> in the sidebar and
<b>Create Credentials</b> for an<b>API-Key</b>.

<b>Copy the API-Key to Lollypop</b>"""),
              "uri_label": "",
              "uri": None,
              "right_button_label": _("Back"),
              "right_button_style": "",
              "left_button_label": _("Finish"),
              "left_button_style": "suggested-action"
            },
        ]
        Assistant.__init__(self, rules)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2019 Jordi Romera <jordiromera@users.sourceforge.net>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, GObject, Gio, Gtk

from gi.repository.Gio import FILE_ATTRIBUTE_STANDARD_NAME, \
                              FILE_ATTRIBUTE_STANDARD_TYPE, \
                              FILE_ATTRIBUTE_STANDARD_IS_HIDDEN,\
                              FILE_ATTRIBUTE_STANDARD_IS_SYMLINK,\
                              FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET,\
                              FILE_ATTRIBUTE_TIME_MODIFIED,\
                              FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE

from gettext import gettext as _
from time import time, sleep
from urllib.parse import urlparse
from multiprocessing import cpu_count

from lollypop.collection_item import CollectionItem
from lollypop.inotify import Inotify
from lollypop.define import App, ScanType, Type, StorageType, ScanUpdate
from lollypop.define import FileType
from lollypop.sqlcursor import SqlCursor
from lollypop.tagreader import TagReader, Discoverer
from lollypop.logger import Logger
from lollypop.database_history import History
from lollypop.objects_track import Track
from lollypop.utils_file import is_audio, is_pls, get_mtime, get_file_type
from lollypop.utils_album import tracks_to_albums
from lollypop.utils import emit_signal, profile, split_list
from lollypop.utils import get_lollypop_album_id, get_lollypop_track_id


SCAN_QUERY_INFO = "{},{},{},{},{},{}".format(
                                       FILE_ATTRIBUTE_STANDARD_NAME,
                                       FILE_ATTRIBUTE_STANDARD_TYPE,
                                       FILE_ATTRIBUTE_STANDARD_IS_HIDDEN,
                                       FILE_ATTRIBUTE_STANDARD_IS_SYMLINK,
                                       FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET,
                                       FILE_ATTRIBUTE_TIME_MODIFIED)


class CollectionScanner(GObject.GObject, TagReader):
    """
        Scan user music collection
    """
    __gsignals__ = {
        "scan-finished": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "updated": (GObject.SignalFlags.RUN_FIRST, None,
                    (GObject.TYPE_PYOBJECT, int))
    }

    def __init__(self):
        """
            Init collection scanner
        """
        GObject.GObject.__init__(self)
        self.__thread = None
        self.__tags = {}
        self.__items = []
        self.__notified_ids = []
        self.__pending_new_artist_ids = []
        self.__history = History()
        self.__progress_total = 1
        self.__progress_count = 0
        self.__progress_fraction = 0
        self.__disable_compilations = not App().settings.get_value(
                "show-compilations")
        if App().settings.get_value("auto-update"):
            self.__inotify = Inotify()
        else:
            self.__inotify = None
        App().albums.update_max_count()

    def update(self, scan_type, uris=[]):
        """
            Update database
            @param scan_type as ScanType
            @param uris as [str]
        """
        self.__disable_compilations = not App().settings.get_value(
                "show-compilations")
        App().lookup_action("update_db").set_enabled(False)
        # Stop previous scan
        if self.is_locked() and scan_type != ScanType.EXTERNAL:
            self.stop()
            GLib.timeout_add(250, self.update, scan_type, uris)
            return
        elif App().ws_director.collection_ws is not None and\
                not App().ws_director.collection_ws.stop():
            GLib.timeout_add(250, self.update, scan_type, uris)
            return
        else:
            if scan_type == ScanType.FULL:
                uris = App().settings.get_music_uris()
            if not uris:
                return
            # Register to progressbar
            if scan_type != ScanType.EXTERNAL:
                App().window.container.progress.add(self)
                App().window.container.progress.set_fraction(0, self)
            Logger.info("Scan started")
            # Launch scan in a separate thread
            self.__thread = App().task_helper.run(self.__scan, scan_type, uris)

    def save_album(self, item):
        """
            Add album to DB
            @param item as CollectionItem
        """
        Logger.debug("CollectionScanner::save_album(): "
                     "Add album artists %s" % item.album_artists)
        (item.new_album_artist_ids,
         item.album_artist_ids) = self.add_artists(item.album_artists,
                                                   item.aa_sortnames,
                                                   item.mb_album_artist_id)
        # We handle artists already created by any previous save_track()
        for artist_id in item.album_artist_ids:
            if artist_id in self.__pending_new_artist_ids:
                item.new_album_artist_ids.append(artist_id)
                self.__pending_new_artist_ids.remove(artist_id)

        item.lp_album_id = get_lollypop_album_id(item.album_name,
                                                 item.album_artists)
        Logger.debug("CollectionScanner::save_track(): Add album: "
                     "%s, %s" % (item.album_name, item.album_artist_ids))
        (item.new_album, item.album_id) = self.add_album(
                                               item.album_name,
                                               item.mb_album_id,
                                               item.lp_album_id,
                                               item.album_artist_ids,
                                               item.uri,
                                               item.album_loved,
                                               item.album_pop,
                                               item.album_rate,
                                               item.album_synced,
                                               item.album_mtime,
                                               item.storage_type)
        if item.year is not None:
            App().albums.set_year(item.album_id, item.year)
            App().albums.set_timestamp(item.album_id, item.timestamp)

    def save_track(self, item):
        """
            Add track to DB
            @param item as CollectionItem
        """
        Logger.debug(
            "CollectionScanner::save_track(): Add artists %s" % item.artists)
        (item.new_artist_ids,
         item.artist_ids) = self.add_artists(item.artists,
                                             item.a_sortnames,
                                             item.mb_artist_id)

        self.__pending_new_artist_ids += item.new_artist_ids
        missing_artist_ids = list(
            set(item.album_artist_ids) - set(item.artist_ids))
        # Special case for broken tags
        # If all artist album tags are missing
        # Can't do more because don't want to break split album behaviour
        if len(missing_artist_ids) == len(item.album_artist_ids):
            item.artist_ids += missing_artist_ids

        if item.genres is None:
            (item.new_genre_ids, item.genre_ids) = ([], [Type.WEB])
        else:
            (item.new_genre_ids, item.genre_ids) = self.add_genres(item.genres)

        item.lp_track_id = get_lollypop_track_id(item.track_name,
                                                 item.artists,
                                                 item.album_name)

        # Add track to db
        Logger.debug("CollectionScanner::save_track(): Add track")
        item.track_id = App().tracks.add(item.track_name,
                                         item.uri,
                                         item.duration,
                                         item.tracknumber,
                                         item.discnumber,
                                         item.discname,
                                         item.album_id,
                                         item.original_year,
                                         item.original_timestamp,
                                         item.track_pop,
                                         item.track_rate,
                                         item.track_loved,
                                         item.track_ltime,
                                         item.track_mtime,
                                         item.mb_track_id,
                                         item.lp_track_id,
                                         item.bpm,
                                         item.storage_type)
        Logger.debug("CollectionScanner::save_track(): Update track")
        self.update_track(item)
        Logger.debug("CollectionScanner::save_track(): Update album")
        self.update_album(item)

    def update_album(self, item):
        """
            Update album artists based on album-artist and artist tags
            This code auto handle compilations: empty "album artist" with
            different artists
            @param item as CollectionItem
        """
        if item.album_artist_ids:
            App().albums.set_artist_ids(item.album_id, item.album_artist_ids)
        # Set artist ids based on content
        else:
            if item.compilation:
                new_album_artist_ids = [Type.COMPILATIONS]
            else:
                new_album_artist_ids = App().albums.calculate_artist_ids(
                    item.album_id, self.__disable_compilations)
            App().albums.set_artist_ids(item.album_id, new_album_artist_ids)
            # We handle artists already created by any previous save_track()
            item.new_album_artist_ids = []
            for artist_id in new_album_artist_ids:
                if artist_id in self.__pending_new_artist_ids:
                    item.new_album_artist_ids.append(artist_id)
                    self.__pending_new_artist_ids.remove(artist_id)
        # Update lp_album_id
        lp_album_id = get_lollypop_album_id(item.album_name,
                                            item.album_artists)
        if lp_album_id != item.lp_album_id:
            App().album_art.move(item.lp_album_id, lp_album_id)
            App().albums.set_lp_album_id(item.album_id, lp_album_id)
            item.lp_album_id = lp_album_id
        # Update album genres
        for genre_id in item.genre_ids:
            App().albums.add_genre(item.album_id, genre_id)
        App().cache.clear_durations(item.album_id)

    def update_track(self, item):
        """
            Set track artists/genres
            @param item as CollectionItem
        """
        # Set artists/genres for track
        for artist_id in item.artist_ids:
            App().tracks.add_artist(item.track_id, artist_id)
        for genre_id in item.genre_ids:
            App().tracks.add_genre(item.track_id, genre_id)

    def del_from_db(self, uri, backup):
        """
            Delete track from db
            @param uri as str
            @param backup as bool
            @return (popularity, ltime, mtime,
                     loved album, album_popularity, album_rate)
        """
        try:
            track_id = App().tracks.get_id_by_uri(uri)
            duration = App().tracks.get_duration(track_id)
            album_id = App().tracks.get_album_id(track_id)
            album_artist_ids = App().albums.get_artist_ids(album_id)
            artist_ids = App().tracks.get_artist_ids(track_id)
            track_pop = App().tracks.get_popularity(track_id)
            track_rate = App().tracks.get_rate(track_id)
            track_ltime = App().tracks.get_ltime(track_id)
            album_mtime = App().tracks.get_mtime(track_id)
            track_loved = App().tracks.get_loved(track_id)
            album_pop = App().albums.get_popularity(album_id)
            album_rate = App().albums.get_rate(album_id)
            album_loved = App().albums.get_loved(album_id)
            album_synced = App().albums.get_synced(album_id)
            if backup:
                f = Gio.File.new_for_uri(uri)
                name = f.get_basename()
                self.__history.add(name, duration, track_pop, track_rate,
                                   track_ltime, album_mtime, track_loved,
                                   album_loved, album_pop, album_rate,
                                   album_synced)
            App().tracks.remove(track_id)
            genre_ids = App().tracks.get_genre_ids(track_id)
            App().albums.clean()
            App().genres.clean()
            App().artists.clean()
            App().cache.clear_durations(album_id)
            SqlCursor.commit(App().db)
            item = CollectionItem(album_id=album_id)
            if not App().albums.get_name(album_id):
                item.artist_ids = []
                for artist_id in album_artist_ids + artist_ids:
                    if not App().artists.get_name(artist_id):
                        item.artist_ids.append(artist_id)
                item.genre_ids = []
                for genre_id in genre_ids:
                    if not App().genres.get_name(genre_id):
                        item.genre_ids.append(genre_id)
                emit_signal(self, "updated", item, ScanUpdate.REMOVED)
            else:
                # Force genre for album
                genre_ids = App().tracks.get_album_genre_ids(album_id)
                App().albums.set_genre_ids(album_id, genre_ids)
                emit_signal(self, "updated", item, ScanUpdate.MODIFIED)
            return (track_pop, track_rate, track_ltime, album_mtime,
                    track_loved, album_loved, album_pop, album_rate)
        except Exception as e:
            Logger.error("CollectionScanner::del_from_db: %s" % e)
        return (0, 0, 0, 0, False, False, 0, 0)

    def is_locked(self):
        """
            True if db locked
            @return bool
        """
        return self.__thread is not None and self.__thread.is_alive()

    def stop(self):
        """
            Stop scan
        """
        self.__thread = None

    def reset_database(self):
        """
            Reset database
        """
        from lollypop.app_notification import AppNotification
        App().window.container.progress.add(self)
        App().window.container.progress.set_fraction(0, self)
        self.__progress_fraction = 0
        notification = AppNotification(_("Resetting database"), [], [], 10000)
        notification.show()
        App().window.container.add_overlay(notification)
        notification.set_reveal_child(True)
        App().task_helper.run(self.__reset_database)

    @property
    def inotify(self):
        """
            Get Inotify object
            @return Inotify
        """
        return self.__inotify

#######################
# PRIVATE             #
#######################
    def __reset_database(self):
        """
            Reset database
        """
        def update_ui():
            App().window.container.go_home()
            App().scanner.update(ScanType.FULL)
        App().player.stop()
        if App().ws_director.collection_ws is not None:
            App().ws_director.collection_ws.stop()
        uris = App().tracks.get_uris()
        i = 0
        SqlCursor.add(App().db)
        SqlCursor.add(self.__history)
        count = len(uris)
        for uri in uris:
            self.del_from_db(uri, True)
            self.__update_progress(i, count, 0.01)
            i += 1
        App().tracks.del_persistent(False)
        App().tracks.clean(False)
        App().albums.clean(False)
        App().artists.clean(False)
        App().genres.clean(False)
        App().cache.clear_table("duration")
        SqlCursor.commit(App().db)
        SqlCursor.remove(App().db)
        SqlCursor.commit(self.__history)
        SqlCursor.remove(self.__history)
        GLib.idle_add(update_ui)

    def __update_progress(self, current, total, allowed_diff):
        """
            Update progress bar status
            @param current as int
            @param total as int
            @param allowed_diff as float => allows to prevent
                                            main loop flooding
        """
        new_fraction = current / total
        if new_fraction > self.__progress_fraction + allowed_diff:
            self.__progress_fraction = new_fraction
            GLib.idle_add(App().window.container.progress.set_fraction,
                          new_fraction, self)

    def __finish(self, items):
        """
            Notify from main thread when scan finished
            @param items as [CollectionItem]
        """
        track_ids = [item.track_id for item in items]
        self.__thread = None
        Logger.info("Scan finished")
        App().lookup_action("update_db").set_enabled(True)
        App().window.container.progress.set_fraction(1.0, self)
        self.stop()
        emit_signal(self, "scan-finished", track_ids)
        # Update max count value
        App().albums.update_max_count()
        # Update featuring
        App().artists.update_featuring()
        if App().ws_director.collection_ws is not None:
            App().ws_director.collection_ws.start()

    def __add_monitor(self, dirs):
        """
            Monitor any change in a list of directory
            @param dirs as str or list of directory to be monitored
        """
        if self.__inotify is None:
            return
        # Add monitors on dirs
        for d in dirs:
            # Handle a stop request
            if self.__thread is None:
                break
            if d.startswith("file://"):
                self.__inotify.add_monitor(d)

    @profile
    def __get_objects_for_uris(self, scan_type, uris):
        """
            Get all tracks and dirs in uris
            @param scan_type as ScanType
            @param uris as string
            @return ([(int, str)], [str], [str])
                    ([(mtime, file)], [dir], [stream])
        """
        files = []
        dirs = []
        streams = []
        walk_uris = []
        # Check collection exists
        for uri in uris:
            parsed = urlparse(uri)
            if parsed.scheme in ["http", "https"]:
                streams.append(uri)
            else:
                f = Gio.File.new_for_uri(uri)
                if f.query_exists():
                    walk_uris.append(uri)
                else:
                    return ([], [], [])

        while walk_uris:
            uri = walk_uris.pop(0)
            try:
                # Directly add files, walk through directories
                f = Gio.File.new_for_uri(uri)
                info = f.query_info(SCAN_QUERY_INFO,
                                    Gio.FileQueryInfoFlags.NONE,
                                    None)
                if info.get_file_type() == Gio.FileType.DIRECTORY:
                    dirs.append(uri)
                    infos = f.enumerate_children(SCAN_QUERY_INFO,
                                                 Gio.FileQueryInfoFlags.NONE,
                                                 None)
                    for info in infos:
                        f = infos.get_child(info)
                        child_uri = f.get_uri()
                        if info.get_is_hidden():
                            continue
                        # User do not want internal symlinks
                        elif info.get_is_symlink() and\
                                App().settings.get_value("ignore-symlinks"):
                            continue
                        walk_uris.append(child_uri)
                    infos.close(None)
                # Only happens if files passed as args
                else:
                    mtime = get_mtime(info)
                    files.append((mtime, uri))
            except Exception as e:
                Logger.error("CollectionScanner::__get_objects_for_uris(): %s"
                             % e)
        files.sort(reverse=True)
        return (files, dirs, streams)

    @profile
    def __scan(self, scan_type, uris):
        """
            Scan music collection for music files
            @param scan_type as ScanType
            @param uris as [str]
            @thread safe
        """
        try:
            self.__items = []
            App().art.clean_rounded()
            (files, dirs, streams) = self.__get_objects_for_uris(
                scan_type, uris)
            if len(uris) != len(streams) and not files:
                self.__flatpak_migration()
                App().notify.send("Lollypop",
                                  _("Scan disabled, missing collection"))
                return
            if scan_type == ScanType.NEW_FILES:
                db_uris = App().tracks.get_uris(uris)
            else:
                db_uris = App().tracks.get_uris()

            # Get mtime of all tracks to detect which has to be updated
            db_mtimes = App().tracks.get_mtimes()
            # * 2 => Scan + Save
            self.__progress_total = len(files) * 2 + len(streams)
            self.__progress_count = 0
            self.__progress_fraction = 0
            # Min: 1 thread, Max: 5 threads
            count = max(1, min(5, cpu_count() // 2))
            split_files = split_list(files, count)
            self.__tags = {}
            self.__notified_ids = []
            self.__pending_new_artist_ids = []
            threads = []
            for files in split_files:
                thread = App().task_helper.run(self.__scan_files,
                                               files, db_mtimes,
                                               scan_type)
                threads.append(thread)
            while threads:
                sleep(0.1)
                thread = threads[0]
                if not thread.is_alive():
                    threads.remove(thread)

            SqlCursor.add(App().db)
            if scan_type == ScanType.EXTERNAL:
                storage_type = StorageType.EXTERNAL
            else:
                storage_type = StorageType.COLLECTION
            self.__items += self.__save_in_db(storage_type)
            # Add streams to DB, only happening on command line/m3u files
            self.__items += self.__save_streams_in_db(streams, storage_type)

            self.__remove_old_tracks(db_uris, scan_type)

            if scan_type == ScanType.EXTERNAL:
                albums = tracks_to_albums(
                    [Track(item.track_id) for item in self.__items])
                App().player.play_albums(albums)
            else:
                self.__add_monitor(dirs)
                GLib.idle_add(self.__finish, self.__items)
            self.__tags = {}
            self.__items = []
            self.__pending_new_artist_ids = []
        except Exception as e:
            Logger.warning("CollectionScanner::__scan(): %s", e)
        SqlCursor.remove(App().db)
        App().settings.set_value("flatpak-access-migration",
                                 GLib.Variant("b", True))

    def __scan_to_handle(self, uri):
        """
            Check if file has to be handle by scanner
            @param f as Gio.File
            @return bool
        """
        try:
            file_type = get_file_type(uri)
            # Get file type using Gio (slower)
            if file_type == FileType.UNKNOWN:
                f = Gio.File.new_for_uri(uri)
                info = f.query_info(FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
                                    Gio.FileQueryInfoFlags.NONE)
                if is_pls(info):
                    file_type = FileType.PLS
                elif is_audio(info):
                    file_type = FileType.AUDIO
            if file_type == FileType.PLS:
                Logger.debug("Importing playlist %s" % uri)
                if App().settings.get_value("import-playlists"):
                    App().playlists.import_tracks(uri)
            elif file_type == FileType.AUDIO:
                Logger.debug("Importing audio %s" % uri)
                return True
        except Exception as e:
            Logger.error("CollectionScanner::__scan_to_handle(): %s" % e)
        return False

    def __scan_files(self, files, db_mtimes, scan_type):
        """
            Scan music collection for new audio files
            @param files as [str]
            @param db_mtimes as {}
            @param scan_type as ScanType
            @thread safe
        """
        discoverer = Discoverer()
        try:
            # Scan new files
            for (mtime, uri) in files:
                # Handle a stop request
                if self.__thread is None and scan_type != ScanType.EXTERNAL:
                    raise Exception("cancelled")
                try:
                    if not self.__scan_to_handle(uri):
                        self.__progress_count += 2
                        continue
                    db_mtime = db_mtimes.get(uri, 0)
                    if mtime > db_mtime:
                        # Do not use mtime if not intial scan
                        if db_mtimes:
                            mtime = int(time())
                        self.__tags[uri] = self.__get_tags(discoverer,
                                                           uri, mtime)
                        self.__progress_count += 1
                        self.__update_progress(self.__progress_count,
                                               self.__progress_total,
                                               0.001)
                    else:
                        # We want to play files, so put them in items
                        if scan_type == ScanType.EXTERNAL:
                            track_id = App().tracks.get_id_by_uri(uri)
                            item = CollectionItem(track_id=track_id)
                            self.__items.append(item)
                        self.__progress_count += 2
                        self.__update_progress(self.__progress_count,
                                               self.__progress_total,
                                               0.1)
                except Exception as e:
                    Logger.error("Scanning file: %s, %s" % (uri, e))
        except Exception as e:
            Logger.warning("CollectionScanner::__scan_files(): % s" % e)

    def __save_in_db(self, storage_type):
        """
            Save current tags into DB
            @param storage_type as StorageType
            @return [CollectionItem]
        """
        items = []
        for uri in list(self.__tags.keys()):
            # Handle a stop request
            if self.__thread is None:
                raise Exception("cancelled")
            Logger.debug("Adding file: %s" % uri)
            tags = self.__tags[uri]
            item = self.__add2db(uri, *tags, storage_type)
            items.append(item)
            self.__progress_count += 1
            self.__update_progress(self.__progress_count,
                                   self.__progress_total,
                                   0.001)
            if item.album_id not in self.__notified_ids:
                self.__notified_ids.append(item.album_id)
                self.__notify_ui(item)
            del self.__tags[uri]
        # Handle a stop request
        if self.__thread is None:
            raise Exception("cancelled")
        return items

    def __save_streams_in_db(self, streams, storage_type):
        """
            Save http stream to DB
            @param streams as [str]
            @param storage_type as StorageType
            @return [CollectionItem]
        """
        items = []
        for uri in streams:
            parsed = urlparse(uri)
            item = self.__add2db(uri, parsed.path, parsed.netloc,
                                 None, "", "", parsed.netloc,
                                 parsed.netloc, "", False, 0, False, 0, 0, 0,
                                 None, 0, "", "", "", "", 1, 0, 0, 0, 0, 0,
                                 False, 0, False, storage_type)
            items.append(item)
            self.__progress_count += 1
        return items

    def __notify_ui(self, item):
        """
            Notify UI for item
            @param items as CollectionItem
        """
        SqlCursor.commit(App().db)
        if item.new_album:
            emit_signal(self, "updated", item, ScanUpdate.ADDED)
        else:
            emit_signal(self, "updated", item, ScanUpdate.MODIFIED)

    def __remove_old_tracks(self, uris, scan_type):
        """
            Remove non existent tracks from DB
            @param scan_type as ScanType
        """
        if scan_type != ScanType.EXTERNAL and self.__thread is not None:
            # We need to check files are always in collections
            if scan_type == ScanType.FULL:
                collections = App().settings.get_music_uris()
            else:
                collections = None
            for uri in uris:
                # Handle a stop request
                if self.__thread is None:
                    raise Exception("cancelled")
                in_collection = True
                if collections is not None:
                    in_collection = False
                    for collection in collections:
                        if collection in uri:
                            in_collection = True
                            break
                f = Gio.File.new_for_uri(uri)
                if not in_collection:
                    Logger.warning(
                        "Removed, not in collection anymore: %s -> %s",
                        uri, collections)
                    self.del_from_db(uri, True)
                elif not f.query_exists():
                    Logger.warning("Removed, file has been deleted: %s", uri)
                    self.del_from_db(uri, True)

    def __get_tags(self, discoverer, uri, track_mtime):
        """
            Read track tags
            @param discoverer as Discoverer
            @param uri as string
            @param track_mtime as int
            @return ()
        """
        f = Gio.File.new_for_uri(uri)
        info = discoverer.get_info(uri)
        tags = info.get_tags()
        name = f.get_basename()
        duration = int(info.get_duration() / 1000000)
        Logger.debug("CollectionScanner::add2db(): Restore stats")
        # Restore stats
        track_id = App().tracks.get_id_by_uri(uri)
        if track_id is None:
            track_id = App().tracks.get_id_by_basename_duration(name,
                                                                duration)
        if track_id is None:
            (track_pop, track_rate, track_ltime,
             album_mtime, track_loved, album_loved,
             album_pop, album_rate, album_synced) = self.__history.get(
                name, duration)
        # Delete track and restore from it
        else:
            (track_pop, track_rate, track_ltime,
             album_mtime, track_loved, album_loved,
             album_pop, album_rate) = self.del_from_db(uri, False)

        Logger.debug("CollectionScanner::add2db(): Read tags")
        title = self.get_title(tags, name)
        version = self.get_version(tags)
        if version != "":
            title += " (%s)" % version
        artists = self.get_artists(tags)
        a_sortnames = self.get_artist_sortnames(tags)
        aa_sortnames = self.get_album_artist_sortnames(tags)
        album_artists = self.get_album_artists(tags)
        album_name = self.get_album_name(tags)
        album_synced = 0
        mb_album_id = self.get_mb_album_id(tags)
        mb_track_id = self.get_mb_track_id(tags)
        mb_artist_id = self.get_mb_artist_id(tags)
        mb_album_artist_id = self.get_mb_album_artist_id(tags)
        genres = self.get_genres(tags)
        discnumber = self.get_discnumber(tags)
        discname = self.get_discname(tags)
        tracknumber = self.get_tracknumber(tags, name)
        # We have popm in tags, override history one
        tag_track_rate = self.get_popm(tags)
        if tag_track_rate > 0:
            track_rate = tag_track_rate
        if album_mtime == 0:
            album_mtime = track_mtime
        bpm = self.get_bpm(tags)
        compilation = self.get_compilation(tags)
        (original_year, original_timestamp) = self.get_original_year(tags)
        (year, timestamp) = self.get_year(tags)
        if year is None:
            (year, timestamp) = (original_year, original_timestamp)
        elif original_year is None:
            (original_year, original_timestamp) = (year, timestamp)
        # If no artists tag, use album artist
        if artists == "":
            artists = album_artists
        if App().settings.get_value("import-advanced-artist-tags"):
            composers = self.get_composers(tags)
            conductors = self.get_conductors(tags)
            performers = self.get_performers(tags)
            remixers = self.get_remixers(tags)
            artists += ";%s" % performers if performers != "" else ""
            artists += ";%s" % conductors if conductors != "" else ""
            artists += ";%s" % composers if composers != "" else ""
            artists += ";%s" % remixers if remixers != "" else ""
        if artists == "":
            artists = _("Unknown")
        return (title, artists, genres, a_sortnames, aa_sortnames,
                album_artists, album_name, discname, album_loved, album_mtime,
                album_synced, album_rate, album_pop, discnumber, year,
                timestamp, original_year, original_timestamp,
                mb_album_id, mb_track_id, mb_artist_id,
                mb_album_artist_id, tracknumber, track_pop, track_rate, bpm,
                track_mtime, track_ltime, track_loved, duration, compilation)

    def __add2db(self, uri, name, artists,
                 genres, a_sortnames, aa_sortnames, album_artists, album_name,
                 discname, album_loved, album_mtime, album_synced, album_rate,
                 album_pop, discnumber, year, timestamp,
                 original_year, original_timestamp, mb_album_id,
                 mb_track_id, mb_artist_id, mb_album_artist_id,
                 tracknumber, track_pop, track_rate, bpm, track_mtime,
                 track_ltime, track_loved, duration, compilation,
                 storage_type=StorageType.COLLECTION):
        """
            Add new file to DB
            @param uri as str
            @param tags as *()
            @param storage_type as StorageType
            @return CollectionItem
        """
        item = CollectionItem(uri=uri,
                              track_name=name,
                              artists=artists,
                              genres=genres,
                              a_sortnames=a_sortnames,
                              aa_sortnames=aa_sortnames,
                              album_artists=album_artists,
                              album_name=album_name,
                              discname=discname,
                              album_loved=album_loved,
                              album_mtime=album_mtime,
                              album_synced=album_synced,
                              album_rate=album_rate,
                              album_pop=album_pop,
                              discnumber=discnumber,
                              year=year,
                              timestamp=timestamp,
                              original_year=original_year,
                              original_timestamp=original_timestamp,
                              mb_album_id=mb_album_id,
                              mb_track_id=mb_track_id,
                              mb_artist_id=mb_artist_id,
                              mb_album_artist_id=mb_album_artist_id,
                              tracknumber=tracknumber,
                              track_pop=track_pop,
                              track_rate=track_rate,
                              bpm=bpm,
                              track_mtime=track_mtime,
                              track_ltime=track_ltime,
                              track_loved=track_loved,
                              duration=duration,
                              compilation=compilation,
                              storage_type=storage_type)
        self.save_album(item)
        self.save_track(item)
        return item

    def __flatpak_migration(self):
        """
            https://github.com/flathub/org.gnome.Lollypop/pull/108
        """
        if GLib.file_test("/app", GLib.FileTest.EXISTS) and\
                not App().settings.get_value("flatpak-access-migration"):
            from lollypop.assistant_flatpak import FlatpakAssistant
            assistant = FlatpakAssistant()
            assistant.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
            assistant.set_transient_for(App().window)
            GLib.timeout_add(1000, assistant.show)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gst, Gtk

from lollypop.define import App, MARGIN_SMALL
from lollypop.utils import ms_to_string
from lollypop.helper_signals import SignalsHelper, signals


class ProgressPlayerWidget(Gtk.Box, SignalsHelper):
    """
        Progress widget synced with player state
    """

    @signals
    def __init__(self):
        """
            Init box
        """
        Gtk.Box.__init__(self)
        # Prevent updating progress while seeking
        self.__seeking = False
        # Update pogress position
        self.__timeout_id = None
        self.__time_label = Gtk.Label.new()
        self.__time_label.show()
        self.__progress = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, None)
        self.__progress.show()
        self.__progress.set_hexpand(True)
        self.__progress.set_draw_value(False)
        self.__progress.connect("change-value", self.__on_change_value)
        self.__multi_press = Gtk.GestureMultiPress.new(self.__progress)
        self.__multi_press.set_propagation_phase(Gtk.PropagationPhase.TARGET)
        self.__multi_press.connect("pressed", self.__on_multi_pressed)
        self.__multi_press.connect("released", self.__on_multi_released)
        self.__multi_press.set_button(1)
        self.__event_controller = Gtk.EventControllerScroll.new(
            self.__progress, Gtk.EventControllerScrollFlags.BOTH_AXES)
        self.__event_controller.set_propagation_phase(
            Gtk.PropagationPhase.TARGET)
        self.__event_controller.connect("scroll", self.__on_scroll)
        self.__total_time_label = Gtk.Label.new()
        self.__total_time_label.show()
        self.set_spacing(MARGIN_SMALL)
        self.pack_start(self.__time_label, False, False, 0)
        self.pack_start(self.__progress, False, True, 0)
        self.pack_start(self.__total_time_label, False, False, 0)
        self.connect("destroy", self.__on_destroy)
        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().player, "status-changed", "_on_status_changed"),
            (App().player, "duration-changed", "_on_duration_changed"),
            (App().player, "seeked", "_on_seeked")
        ]

    def update(self):
        """
            Update progress state
        """
        player = App().player
        self._on_current_changed(player)
        self._on_status_changed(player)
        self.update_position()

    def update_position(self, value=None):
        """
            Update progress bar position
            @param value as int
        """
        if not self.__seeking:
            if value is None and App().player.get_status() != Gst.State.PAUSED:
                value = App().player.position
            if value is not None and value >= 0:
                self.__progress.set_value(value)
                time_string = ms_to_string(value)
                self.__time_label.set_markup(
                    "<span font_features='tnum'>%s</span>" % time_string)
        return True

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update scale on current changed
            @param player as Player
        """
        style_context = self.__progress.get_style_context()
        style_context.remove_class("youtube-scale")
        if App().player.current_track.id is None:
            self.__total_time_label.set_text("")
            self.__time_label.set_text("")
            return

        self.__progress.set_value(0.0)
        self.__time_label.set_markup("<span font_features='tnum'>0:00</span>")
        if App().player.current_track.is_web:
            style_context.add_class("youtube-scale")
        self.__progress.set_range(0.0,
                                  App().player.current_track.duration)
        time_string = ms_to_string(App().player.current_track.duration)
        self.__total_time_label.set_markup(
            "<span font_features='tnum'>%s</span>" % time_string)

    def _on_duration_changed(self, player, track_id):
        """
            Update duration
            @param player as Player
            @param track_id as int
        """
        if track_id == player.current_track.id:
            duration = player.current_track.duration
            self.__progress.set_range(0.0, duration)
            time_string = ms_to_string(duration)
            self.__total_time_label.set_markup(
                "<span font_features='tnum'>%s</span>" % time_string)

    def _on_status_changed(self, player):
        """
            Update buttons and progress bar
            @param player as Player
        """
        if player.is_playing:
            if self.__timeout_id is None:
                self.__timeout_id = GLib.timeout_add(1000,
                                                     self.update_position)
        else:
            if self.__timeout_id is not None:
                GLib.source_remove(self.__timeout_id)
                self.__timeout_id = None

    def _on_seeked(self, player, position):
        """
            Update position
            @param position as int
        """
        self.update_position(position)

#######################
# PRIVATE             #
#######################
    def __on_change_value(self, scale, scroll_type, value):
        """
            Update label
            @param scale as Gtk.Scale
            @param scroll_type as Gtk.ScrollType
            @param value as float
        """
        value = min(value, scale.get_adjustment().get_upper())
        time_string = ms_to_string(value)
        self.__time_label.set_markup(
            "<span font_features='tnum'>%s</span>" % time_string)

    def __on_multi_pressed(self, gesture, n_press, x, y):
        """
            On press, mark player as seeking
            @param gesture as Gtk.Gesture
            @param n_press as int
            @param x as int
            @param y as int
        """
        self.__seeking = True

    def __on_multi_released(self, gesture, n_press, x, y):
        """
            Callback for scale release button
            @param gesture as Gtk.Gesture
            @param n_press as int
            @param x as int
            @param y as int
        """
        if n_press != 1:
            return
        value = self.__progress.get_value()
        App().player.seek(value)
        self.__seeking = False
        self.update_position(value)

    def __on_scroll(self, event_controler, x, y):
        """
            Seek forward or backward
            @param event_controller as Gtk.EventControllerScroll
            @param x as int
            @param y as int
        """
        if x == 0:
            diff = -y
        else:
            diff = -x
        if App().player.is_playing:
            position = App().player.position
            seek = position + diff * 5000
            seek = max(min(App().player.current_track.duration - 2000, seek),
                       0)
            App().player.seek(seek)
            self.update_position(seek)

    def __on_destroy(self, widget):
        """
            Stop timeout
            @param widget as Gtk.Widget
        """
        self.__multi_press = None
        self.__event_controller = None
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

from lollypop.logger import Logger
from lollypop.define import App
from lollypop.objects_album import Album
from lollypop.utils import get_default_storage_type


def play_artists(artist_ids, genre_ids):
    """
        Play artists
        @param artist_ids as [int]
        @param genre_ids as [int]
    """
    try:
        storage_type = get_default_storage_type()
        if App().player.is_party:
            App().lookup_action("party").change_state(
                GLib.Variant("b", False))
        album_ids = App().albums.get_ids(genre_ids, artist_ids, storage_type,
                                         False)
        if App().settings.get_value("play-featured"):
            album_ids += App().artists.get_featured(genre_ids,
                                                    artist_ids,
                                                    storage_type,
                                                    False)
        albums = []
        for album_id in album_ids:
            albums.append(Album(album_id, genre_ids, artist_ids, False))
        App().player.play_albums(albums)
    except Exception as e:
        Logger.error("play_artists(): %s" % e)


def add_artist_to_playback(artist_ids, genre_ids, add):
    """
        Add artist to current playback
        @param artist_ids as [int]
        @param genre_ids as [int]
        @param add as bool
    """
    try:
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_ids(genre_ids, artist_ids, storage_type,
                                         False)
        if App().settings.get_value("play-featured"):
            album_ids += App().artists.get_featured(genre_ids,
                                                    artist_ids,
                                                    storage_type,
                                                    False)
        if add:
            albums = []
            for album_id in album_ids:
                if album_id not in App().player.album_ids:
                    album = Album(album_id, genre_ids, artist_ids, False)
                    albums.append(album)
            App().player.add_albums(albums)
        else:
            App().player.remove_album_by_ids(album_ids)
            if len(App().player.albums) == 0:
                App().player.stop()
            elif App().player.current_track.album.id\
                    not in App().player.album_ids:
                App().player.skip_album()
    except Exception as e:
        Logger.error("add_artist_to_playback(): %s" % e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, GLib, Gio, TotemPlParser

from gettext import gettext as _
import itertools
import sqlite3
from datetime import datetime
from threading import Lock
import json

from lollypop.database import Database
from lollypop.define import App, Type
from lollypop.objects_track import Track
from lollypop.sqlcursor import SqlCursor
from lollypop.localized import LocalizedCollation
from lollypop.shown import ShownPlaylists
from lollypop.utils import emit_signal, get_default_storage_type
from lollypop.utils_file import get_mtime
from lollypop.logger import Logger
from lollypop.database_upgrade import DatabasePlaylistsUpgrade


class Playlists(GObject.GObject):
    """
        Playlists manager
    """
    __LOCAL_PATH = GLib.get_user_data_dir() + "/lollypop"
    _DB_PATH = "%s/playlists.db" % __LOCAL_PATH
    __gsignals__ = {
        "playlists-added": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "playlists-removed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "playlists-updated": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "playlists-renamed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "playlist-track-added": (
            GObject.SignalFlags.RUN_FIRST, None, (int, str)),
        "playlist-track-removed": (
            GObject.SignalFlags.RUN_FIRST, None, (int, str))
    }
    __create_playlists = """CREATE TABLE playlists (
                            id INTEGER PRIMARY KEY,
                            name TEXT NOT NULL,
                            synced INT NOT NULL DEFAULT 0,
                            smart_enabled INT NOT NULL DEFAULT 0,
                            smart_sql TEXT,
                            uri TEXT,
                            mtime BIGINT NOT NULL)"""

    __create_tracks = """CREATE TABLE tracks (
                        playlist_id INT NOT NULL,
                        uri TEXT NOT NULL)"""

    def __init__(self):
        """
            Init playlists manager
        """
        self.thread_lock = Lock()
        GObject.GObject.__init__(self)
        upgrade = DatabasePlaylistsUpgrade()
        # Create db schema
        f = Gio.File.new_for_path(self._DB_PATH)
        if not f.query_exists():
            try:
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_playlists)
                    sql.execute(self.__create_tracks)
                    sql.execute("PRAGMA user_version=%s" % upgrade.version)
            except:
                pass
        else:
            upgrade.upgrade(self)

    def get_new_name(self):
        """
            Get a name for a new playlist
            @return str
        """
        existing_playlists = []
        for (playlist_id, name) in App().playlists.get():
            existing_playlists.append(name)

        # Search for an available name
        count = 1
        name = _("New playlist ") + str(count)
        while name in existing_playlists:
            count += 1
            name = _("New playlist ") + str(count)
        return name

    def add(self, name):
        """
            Add a playlist
            @param name as str
            @return playlist_id as int
            @thread safe
        """
        if name == _("Loved tracks"):
            return Type.LOVED
        lastrowid = 0
        with SqlCursor(self, True) as sql:
            result = sql.execute("INSERT INTO playlists (name, mtime)\
                                  VALUES (?, ?)",
                                 (name, 0))
            lastrowid = result.lastrowid
        emit_signal(self, "playlists-added", lastrowid)
        return lastrowid

    def exists(self, playlist_id):
        """
            Return True if playlist exists
            @param playlist_id as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM playlists\
                                  WHERE rowid=?",
                                 (playlist_id,))
            v = result.fetchone()
            if v is not None:
                return True
            else:
                return False

    def rename(self, playlist_id, name):
        """
            Rename playlist
            @param playlist_id as int
            @param name as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists\
                        SET name=?\
                        WHERE rowid=?",
                        (name, playlist_id))
        emit_signal(self, "playlists-renamed", playlist_id)
        App().art.remove_from_cache("playlist_" + name, "ROUNDED")

    def remove(self, playlist_id):
        """
            Remove playlist
            @param playlist_id as int
        """
        name = self.get_name(playlist_id)
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM playlists\
                        WHERE rowid=?",
                        (playlist_id,))
            sql.execute("DELETE FROM tracks\
                        WHERE playlist_id=?",
                        (playlist_id,))
        emit_signal(self, "playlists-removed", playlist_id)
        App().art.remove_from_cache("playlist_" + name, "ROUNDED")

    def clear(self, playlist_id):
        """
            Clear playlsit
            @param playlist_id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM tracks\
                         WHERE playlist_id=?", (playlist_id,))
        self.sync_to_disk(playlist_id)

    def add_uri(self, playlist_id, uri, signal=False):
        """
            Add uri to playlist
            @param playlist_id as int
            @param uri as str
            @param signal as bool
        """
        if self.exists_track(playlist_id, uri):
            return
        with SqlCursor(self, True) as sql:
            sql.execute("INSERT INTO tracks VALUES (?, ?)", (playlist_id, uri))
        if signal:
            emit_signal(self, "playlist-track-added", playlist_id, uri)

    def add_uris(self, playlist_id, uris, signal=False):
        """
            Add uris to playlists (even if exists)
            @param playlist_id as int
            @param uris as [str]
            @param signal as bool
        """
        for uri in uris:
            self.add_uri(playlist_id, uri, signal)
        self.sync_to_disk(playlist_id)

    def add_tracks(self, playlist_id, tracks, signal=False):
        """
            Add tracks to playlist
            @param playlist_id as int
            @param tracks as [Track]
            @param signal as bool
        """
        for track in tracks:
            self.add_uri(playlist_id, track.uri, signal)
        self.sync_to_disk(playlist_id)

    def remove_uri(self, playlist_id, uri, signal=False):
        """
            Remove uri from playlist
            @param playlist_id as int
            @param uri a str
            @param signal as bool
        """
        if not self.exists_track(playlist_id, uri):
            return
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM tracks WHERE uri=? AND playlist_id=?",
                        (uri, playlist_id))
        if signal:
            emit_signal(self, "playlist-track-removed", playlist_id, uri)

    def remove_uris(self, playlist_id, uris, signal=False):
        """
            Remove uris from playlist
            @param playlist_id as int
            @param uris as [str]
            @param signal as bool
        """
        for uri in uris:
            self.remove_uri(playlist_id, uri, signal)
        self.sync_to_disk(playlist_id)

    def remove_tracks(self, playlist_id, tracks, signal=False):
        """
            Remove tracks from playlist
            @param playlist_id as int
            @param tracks as [Track]
            @param signal as bool
        """
        for track in tracks:
            self.remove_uri(playlist_id, track.uri, signal)
        self.sync_to_disk(playlist_id)

    def get(self):
        """
            Return availables playlists
            @return [int, str, str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid, name\
                                  FROM playlists\
                                  ORDER BY name\
                                  COLLATE NOCASE COLLATE LOCALIZED")
            return list(result)

    def get_ids(self):
        """
            Return availables playlists
            @return [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM playlists\
                                  ORDER BY name\
                                  COLLATE NOCASE COLLATE LOCALIZED")
            return list(itertools.chain(*result))

    def get_track_uris(self, playlist_id):
        """
            Return available track uris for playlist
            @param playlist_id as int
            @return [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri\
                                  FROM tracks\
                                  WHERE playlist_id=?", (playlist_id,))
            return list(itertools.chain(*result))

    def get_smart_track_uris(self, playlist_id):
        """
            Return available track uris for playlist
            @param playlist_id as int
            @return [str]
        """
        request = self.get_smart_sql(playlist_id)
        # We need to inject skipped/storage_type
        storage_type = get_default_storage_type()
        split = request.split("ORDER BY")
        split[0] += " AND tracks.loved != %s" % Type.NONE
        split[0] += " AND tracks.storage_type&%s " % storage_type
        track_ids = App().db.execute("ORDER BY".join(split))
        return [Track(track_id).uri for track_id in track_ids]

    def get_track_ids(self, playlist_id):
        """
            Return availables track ids for playlist
            @param playlist_id as int
            @return [int]
        """
        track_ids = []
        limit = App().settings.get_value("view-limit").get_int32()
        storage_type = get_default_storage_type()
        if playlist_id == Type.POPULARS:
            track_ids = App().tracks.get_populars([], storage_type,
                                                  False, limit)
        elif playlist_id == Type.RECENTS:
            track_ids = App().tracks.get_recently_listened_to(storage_type,
                                                              False,
                                                              limit)
        elif playlist_id == Type.LITTLE:
            track_ids = App().tracks.get_little_played(storage_type,
                                                       False,
                                                       limit)
        elif playlist_id == Type.RANDOMS:
            track_ids = App().tracks.get_randoms([], storage_type,
                                                 False, limit)
        elif playlist_id == Type.SKIPPED:
            track_ids = App().tracks.get_skipped(storage_type)
        elif playlist_id == Type.ALL:
            track_ids = App().tracks.get_ids(storage_type, False)
        elif playlist_id == Type.LOVED:
            track_ids = App().tracks.get_loved_track_ids([], storage_type)
        else:
            with SqlCursor(self) as sql:
                result = sql.execute("SELECT music.tracks.rowid\
                                      FROM tracks, music.tracks\
                                      WHERE tracks.playlist_id=?\
                                      AND music.tracks.uri=\
                                      main.tracks.uri",
                                     (playlist_id,))
                track_ids = list(itertools.chain(*result))
        return track_ids

    def get_tracks(self, playlist_id):
        """
            Return availables tracks for playlist
            @param playlist_id as int
            @return [Track]
        """
        return [Track(track_id)
                for track_id in self.get_track_ids(playlist_id)]

    def get_duration(self, playlist_id):
        """
            Return playlist duration
            @param playlist_id as int
            @return duration as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT SUM(music.tracks.duration)\
                                  FROM tracks, music.tracks\
                                  WHERE tracks.playlist_id=?\
                                  AND music.tracks.uri=\
                                  main.tracks.uri",
                                 (playlist_id,))
            v = result.fetchone()
            if v is not None and v[0] is not None:
                return v[0]
            return 0

    def get_id(self, playlist_name):
        """
            Get playlist id
            @param playlist_name as str
            @return playlst id as int
        """
        if playlist_name == _("Loved tracks"):
            return Type.LOVED

        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                 FROM playlists\
                                 WHERE name=?", (playlist_name,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_name(self, playlist_id):
        """
            Get playlist name
            @param playlist_id as int
            @return playlist name as str
        """
        if playlist_id < 0:
            for (id, name, sortname) in ShownPlaylists.get(True):
                if id == playlist_id:
                    return name

        with SqlCursor(self) as sql:
            result = sql.execute("SELECT name\
                                 FROM playlists\
                                 WHERE rowid=?", (playlist_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_synced(self, playlist_id, index):
        """
            True if playlist synced
            @param playlist_id as int
            @param index as int
            @return bool
        """
        if playlist_id < 0:
            internal_ids = App().settings.get_value(
                    "sync-internal-ids").get_string()
            try:
                data = json.loads(internal_ids)
                data.keys()
            except:
                data = {}
            synced_ids = []
            for synced_id in data.keys():
                if data[synced_id] & (1 << index):
                    synced_ids.append(int(synced_id))
            return playlist_id in synced_ids
        else:
            with SqlCursor(self) as sql:
                result = sql.execute("SELECT synced\
                                     FROM playlists\
                                     WHERE rowid=?", (playlist_id,))
                v = result.fetchone()
                if v is not None:
                    return v[0] & (1 << index)
                return False

    def get_synced_ids(self, index):
        """
            Return availables synced playlists
            @return [int]
        """
        with SqlCursor(self) as sql:
            internal_ids = App().settings.get_value(
                    "sync-internal-ids").get_string()
            try:
                data = json.loads(internal_ids)
                data.keys()
            except:
                data = {}
            synced_ids = []
            for playlist_id in data.keys():
                if data[playlist_id] & (1 << index):
                    synced_ids.append(int(playlist_id))
            result = sql.execute("SELECT rowid\
                                  FROM playlists\
                                  WHERE synced & (1 << ?)\
                                  ORDER BY name\
                                  COLLATE NOCASE COLLATE LOCALIZED",
                                 (index,))
            return list(itertools.chain(*result)) + synced_ids

    def get_smart(self, playlist_id):
        """
            True if playlist is smart
            @param playlist_id as int
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT smart_enabled\
                                 FROM playlists\
                                 WHERE rowid=?", (playlist_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return False

    def get_smart_sql(self, playlist_id):
        """
            Get SQL smart request
            @param playlist_id as int
            @return str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT smart_sql\
                                 FROM playlists\
                                 WHERE rowid=?", (playlist_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def set_synced(self, playlist_id, synced):
        """
            Mark playlist as synced
            @param playlist_id as int
            @param synced as bool
        """
        if playlist_id < 0:
            internal_ids = App().settings.get_value(
                "sync-internal-ids").get_string()
            try:
                data = json.loads(internal_ids)
                data.keys()
            except:
                data = {}
            data[str(playlist_id)] = synced
            internal_ids = json.dumps(data)
            App().settings.set_value("sync-internal-ids",
                                     GLib.Variant("s", internal_ids))
        else:
            with SqlCursor(self, True) as sql:
                sql.execute("UPDATE playlists\
                            SET synced=?\
                            WHERE rowid=?",
                            (synced, playlist_id))

    def set_smart(self, playlist_id, smart):
        """
            Mark playlist as smart
            @param playlist_id as int
            @param smart as bool
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists\
                        SET smart_enabled=?\
                        WHERE rowid=?",
                        (smart, playlist_id))
            emit_signal(self, "playlists-updated", playlist_id)

    def set_smart_sql(self, playlist_id, request):
        """
            Set playlist SQL smart request
            @param playlist_id as int
            @param request as str
        """
        name = self.get_name(playlist_id)
        # Clear cache
        App().art.remove_from_cache("playlist_" + name, "ROUNDED")
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists\
                        SET smart_sql=?\
                        WHERE rowid=?",
                        (request, playlist_id))
            emit_signal(self, "playlists-updated", playlist_id)

    def get_position(self, playlist_id, track_id):
        """
            Get track position in playlist
            @param playlist_id as int
            @param track_id as int
            @return position as int
        """
        i = 0
        for tid in self.get_track_ids(playlist_id):
            if track_id == tid:
                break
            i += 1
        return i

    def exists_track(self, playlist_id, uri):
        """
            Check if track id exist in playlist
            @param playlist_id as int
            @param uri as str
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri\
                                  FROM tracks\
                                  WHERE playlist_id=?\
                                  AND uri=?", (playlist_id, uri))
            v = result.fetchone()
            if v is not None:
                return True
            return False

    def set_sync_uri(self, playlist_id, uri):
        """
            Set sync URI
            @param playlist_id as int
            @param uri as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists\
                        SET uri=?\
                        WHERE rowid=?", (uri, playlist_id))

    def get_sync_uri(self, playlist_id):
        """
            Get sync URI
            @param playlist_id as int
            @return str/None
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri\
                                  FROM playlists\
                                  WHERE rowid=?", (playlist_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def sync_to_disk(self, playlist_id, create=False):
        """
            Sync playlist_id to disk
            @param playlist_id as int
            @param create as bool => create file
        """
        try:
            name = self.get_name(playlist_id)
            # Clear cache
            App().art.remove_from_cache("playlist_" + name, "ROUNDED")
            playlist_uri = self.get_sync_uri(playlist_id)
            if playlist_uri is None:
                return
            f = Gio.File.new_for_uri(playlist_uri)
            if not f.query_exists() and not create:
                return
            if self.get_smart(playlist_id):
                uris = self.get_smart_track_uris(playlist_id)
            else:
                uris = self.get_track_uris(playlist_id)
            if not uris:
                return
            stream = f.replace(None, False,
                               Gio.FileCreateFlags.REPLACE_DESTINATION, None)
            stream.write("#EXTM3U\n".encode("utf-8"))
            playlist_dir_uri = playlist_uri.replace(f.get_basename(), "")
            for uri in uris:
                if uri.startswith("web://"):
                    continue
                if playlist_dir_uri in uri:
                    filepath = uri.replace(playlist_dir_uri, "")
                    string = "%s\n" % GLib.uri_unescape_string(filepath, None)
                else:
                    string = "%s\n" % uri
                stream.write(string.encode("utf-8"))
            stream.close()
        except Exception as e:
            Logger.error("Playlists::sync_to_disk(): %s", e)

    def exists_album(self, playlist_id, album):
        """
            Return True if object_id is already present in playlist
            @param playlist_id as int
            @param album as Album/Disc
            @return bool
        """
        # We do not use Album object for performance reasons
        playlist_uris = self.get_track_uris(playlist_id)
        track_uris = album.track_uris
        return len(set(playlist_uris) & set(track_uris)) == len(track_uris)

    def remove_device(self, index):
        """
            Remove device from DB
            @param index as int => device index
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists SET synced = synced & ~(1<<?)",
                        (index,))

    def import_tracks(self, uri):
        """
            Import file as playlist
            @param uri as str
        """
        f = Gio.File.new_for_uri(uri)
        # Create playlist and get id
        basename = ".".join(f.get_basename().split(".")[:-1])
        playlist_id = self.get_id(basename)
        # Do not reimport playlists
        if playlist_id is not None:
            return
        playlist_id = self.add(basename)
        # Check mtime has been updated
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT mtime\
                                 FROM playlists\
                                 WHERE rowid=?", (playlist_id,))
            v = result.fetchone()
            if v is not None:
                db_mtime = v[0]
            else:
                db_mtime = 0
            info = f.query_info(Gio.FILE_ATTRIBUTE_TIME_MODIFIED,
                                Gio.FileQueryInfoFlags.NONE, None)
            mtime = get_mtime(info)
            if db_mtime >= mtime:
                return

        # Load playlist
        parser = TotemPlParser.Parser.new()
        uris = []
        parser.connect("entry-parsed", self.__on_entry_parsed,
                       playlist_id, uris)
        parser.parse_async(f.get_uri(), True,
                           None, self.__on_parse_finished,
                           playlist_id, uris)

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            sql = sqlite3.connect(self._DB_PATH, 600.0)
            sql.execute('ATTACH DATABASE "%s" AS music' % Database.DB_PATH)
            sql.create_collation("LOCALIZED", LocalizedCollation())
            return sql
        except:
            exit(-1)

#######################
# PRIVATE             #
#######################
    def __on_parse_finished(self, parser, result, playlist_id, uris):
        """
            Add tracks to playlists
            @param parser as TotemPlParser.Parser
            @param result as Gio.AsyncResult
            @param playlist_id as int
            @param uris as [str]
        """
        self.clear(playlist_id)
        self.add_uris(playlist_id, uris)
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE playlists SET mtime=?\
                         WHERE rowid=?", (datetime.now().strftime("%s"),
                                          playlist_id))

    def __on_entry_parsed(self, parser, uri, metadata, playlist_id, uris):
        """
            Play stream
            @param parser as TotemPlParser.Parser
            @param track uri as str
            @param metadata as GLib.HastTable
            @param playlist_id as int
            @param uris as [str]
        """
        uris.append(uri)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GdkPixbuf, GLib, Gst

from random import choice
from gettext import gettext as _
from time import time

from lollypop.helper_task import TaskHelper
from lollypop.tagreader import Discoverer
from lollypop.artwork_manager import ArtworkManager
from lollypop.artwork_downloader_album import AlbumArtworkDownloader
from lollypop.logger import Logger
from lollypop.define import CACHE_PATH, ALBUMS_WEB_PATH, ALBUMS_PATH
from lollypop.define import ArtSize, StorageType
from lollypop.define import App, StoreExtention, ArtBehaviour
from lollypop.utils import emit_signal
from lollypop.utils_file import create_dir, is_readonly


class AlbumArtwork(ArtworkManager, AlbumArtworkDownloader):
    """
        Album artwork manager
    """

    __MIMES = ("jpeg", "jpg", "png", "gif")

    def __init__(self):
        """
            Init album artwork manager
        """
        ArtworkManager.__init__(self)
        AlbumArtworkDownloader.__init__(self)
        create_dir(ALBUMS_PATH)
        create_dir(ALBUMS_WEB_PATH)
        self.__favorite = App().settings.get_value(
            "favorite-cover").get_string()
        if not self.__favorite:
            self.__favorite = App().settings.get_default_value(
                "favorite-cover").get_string()

    def get_cache_path(self, album, width, height):
        """
            get artwork cache path for album_id
            @param album as Album
            @param width as int
            @param height as int
            @return cover path as string or None if no cover
        """
        try:
            cache_path = "%s/%s_%s_%s" % (CACHE_PATH,
                                          album.lp_album_id,
                                          width,
                                          height)
            cache_path = self.add_extension(cache_path)
            f = Gio.File.new_for_path(cache_path)
            if f.query_exists():
                return cache_path
            else:
                self.get(album, width, height, 1)
                if f.query_exists():
                    return cache_path
        except Exception as e:
            Logger.error("AlbumArtwork::get_cache_path(): %s" % e)
        return None

    def get_uri(self, album):
        """
            Look for artwork in dir:
            - favorite from settings first
            - Artist_Album.jpg then
            - Any any supported image otherwise
            @param album as Album
            @return cover uri as string
        """
        if album.id is None:
            return None
        try:
            self.__update_uri(album)
            if not album.storage_type & StorageType.COLLECTION:
                store_path = "%s/%s" % (ALBUMS_WEB_PATH, album.lp_album_id)
                store_path = self.add_extension(store_path)
                uris = [GLib.filename_to_uri(store_path)]
            else:
                store_path = "%s/%s" % (ALBUMS_PATH, album.lp_album_id)
                store_path = self.add_extension(store_path)
                if self.extension == StoreExtention.PNG:
                    uris = [
                        # Default favorite artwork
                        "%s/%s.png" % (album.uri, self.__favorite),
                        "%s/%s.jpg" % (album.uri, self.__favorite),
                        # Used when album.uri is readonly or for Web
                        GLib.filename_to_uri(store_path),
                        # Used when having muliple albums in same folder
                        "%s/%s.png" % (album.uri, album.lp_album_id),
                        "%s/%s.jpg" % (album.uri, album.lp_album_id)
                    ]
                else:
                    uris = [
                        # Default favorite artwork
                        "%s/%s.jpg" % (album.uri, self.__favorite),
                        "%s/%s.png" % (album.uri, self.__favorite),
                        # Used when album.uri is readonly or for Web
                        GLib.filename_to_uri(store_path),
                        # Used when having muliple albums in same folder
                        "%s/%s.jpg" % (album.uri, album.lp_album_id),
                        "%s/%s.png" % (album.uri, album.lp_album_id)
                    ]
            for uri in uris:
                f = Gio.File.new_for_uri(uri)
                if f.query_exists():
                    return uri
        except Exception as e:
            Logger.error("AlbumArtwork::get_uri(): %s", e)
        return None

    def get_uris(self, album):
        """
            Get locally available artworks for album
            @param album as Album
            @return [paths]
        """
        if not album.storage_type & (StorageType.COLLECTION |
                                     StorageType.EXTERNAL):
            return []
        try:
            uris = []
            f = Gio.File.new_for_uri(album.uri)
            infos = f.enumerate_children(
                "standard::name",
                Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                None)
            all_uris = []
            for info in infos:
                f = infos.get_child(info)
                all_uris.append(f.get_uri())
            for uri in filter(lambda p: p.lower().endswith(self.__MIMES),
                              all_uris):
                uris.append(uri)
            infos.close(None)
        except Exception as e:
            Logger.error("AlbumArtwork::get_uris(): %s", e)
        return uris

    def get(self, album, width, height, scale_factor,
            behaviour=ArtBehaviour.CACHE | ArtBehaviour.CROP_SQUARE):
        """
            Return a cairo surface for album_id, covers are cached as jpg.
            @param album as Album
            @param width as int
            @param height as int
            @param scale_factor factor as int
            @param behaviour as ArtBehaviour
            @return cairo surface
            @thread safe
        """
        uri = None
        if album.id is None:
            return None
        width *= scale_factor
        height *= scale_factor
        # Blur when reading from tags can be slow, so prefer cached version
        # Blur allows us to ignore width/height until we want CROP/CACHE
        optimized_blur = behaviour & (ArtBehaviour.BLUR |
                                      ArtBehaviour.BLUR_HARD) and\
            not behaviour & (ArtBehaviour.CACHE |
                             ArtBehaviour.CROP |
                             ArtBehaviour.CROP_SQUARE)
        if optimized_blur:
            w = ArtSize.BIG * scale_factor
            h = ArtSize.BIG * scale_factor
        else:
            w = width
            h = height
        cache_path = "%s/%s_%s_%s" % (CACHE_PATH, album.lp_album_id, w, h)
        cache_path = self.add_extension(cache_path)
        pixbuf = None
        try:
            # Look in cache
            f = Gio.File.new_for_path(cache_path)
            if not behaviour & ArtBehaviour.NO_CACHE and f.query_exists():
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(cache_path)
                if optimized_blur:
                    pixbuf = self.load_behaviour(pixbuf,
                                                 width, height, behaviour)
                return pixbuf
            # Use favorite folder artwork
            if pixbuf is None:
                uri = self.get_uri(album)
                data = None
                if uri is not None:
                    f = Gio.File.new_for_uri(uri)
                    (status, data, tag) = f.load_contents(None)
                    bytes = GLib.Bytes.new(data)
                    stream = Gio.MemoryInputStream.new_from_bytes(bytes)
                    pixbuf = GdkPixbuf.Pixbuf.new_from_stream(
                        stream, None)
                    stream.close()

            # Use tags artwork
            if pixbuf is None and album.tracks and\
                    album.storage_type & (StorageType.COLLECTION |
                                          StorageType.EXTERNAL):
                try:
                    track = choice(album.tracks)
                    pixbuf = self.__get_pixbuf_from_tags(track.uri)
                except Exception as e:
                    Logger.error("AlbumArtwork::get(): %s", e)

            # Use folder artwork
            if pixbuf is None and\
                    album.storage_type & (StorageType.COLLECTION |
                                          StorageType.EXTERNAL):
                uri = self.__get_first(album)
                # Look in album folder
                if uri is not None:
                    f = Gio.File.new_for_uri(uri)
                    (status, data, tag) = f.load_contents(None)
                    bytes = GLib.Bytes.new(data)
                    stream = Gio.MemoryInputStream.new_from_bytes(bytes)
                    pixbuf = GdkPixbuf.Pixbuf.new_from_stream(
                        stream, None)
                    stream.close()
            if pixbuf is None:
                self.download(album.id)
                return None
            pixbuf = self.load_behaviour(pixbuf,
                                         width, height, behaviour)
            if behaviour & ArtBehaviour.CACHE:
                self.save_pixbuf(pixbuf, cache_path)
            return pixbuf
        except Exception as e:
            Logger.warning("AlbumArtwork::get(): %s -> %s" % (uri, e))
            return None

    def add(self, album, data):
        """
            Add artwork for album as data
            @param album as Album
            @param data as bytes
        """
        self.uncache(album)
        try:
            if not album.storage_type & StorageType.COLLECTION:
                self.__save_web(album, data)
            elif is_readonly(album.uri):
                self.__save_ro(album, data)
            else:
                self.__save(album, data)
        except Exception as e:
            Logger.error("AlbumArtwork::add(): %s" % e)

    def move(self, old_lp_album_id, new_lp_album_id):
        """
            Move artwork from an old id to a new id
            @param old_lp_album_id as str
            @param new_lp_album_id s str
        """
        try:
            for store in [ALBUMS_WEB_PATH, ALBUMS_PATH]:
                old_path = "%s/%s" % (store, old_lp_album_id)
                old_path = self.add_extension(old_path)
                old = Gio.File.new_for_path(old_path)
                if old.query_exists():
                    new_path = "%s/%s" % (store, new_lp_album_id)
                    new_path = self.add_extension(new_path)
                    new = Gio.File.new_for_path(new_path)
                    old.move(new, Gio.FileCopyFlags.OVERWRITE, None, None)
                    break
        except Exception as e:
            Logger.error("AlbumArtwork::move(): %s" % e)

    def uncache(self, album, width=-1, height=-1):
        """
            Remove cover from cache for album id
            @param album as Album
            @param width as int
            @param height as int
        """
        try:
            from pathlib import Path
            if width == -1 or height == -1:
                if self.extension == StoreExtention.PNG:
                    extension = "png"
                else:
                    extension = "jpg"
                for p in Path(CACHE_PATH).glob(
                        "%s*.%s" % (album.lp_album_id, extension)):
                    p.unlink()
            else:
                cache_path = "%s/%s_%s_%s" % (CACHE_PATH,
                                              album.lp_album_id,
                                              width,
                                              height)
                f = Gio.File.new_for_path(self.add_extension(cache_path))
                if f.query_exists():
                    f.delete()
        except Exception as e:
            Logger.error("AlbumArtwork::clean(): %s" % e)

#######################
# PRIVATE             #
#######################
    def __get_first(self, album):
        """
            Get first locally available artwork for album
            @param album as Album
            @return path or None
        """
        # Folders with many albums, get_uri()
        if App().albums.get_uri_count(album.uri) > 1:
            return None
        if not album.storage_type & (StorageType.COLLECTION |
                                     StorageType.EXTERNAL):
            return None
        f = Gio.File.new_for_uri(album.uri)
        infos = f.enumerate_children("standard::name",
                                     Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                                     None)
        all_uris = []
        for info in infos:
            f = infos.get_child(info)
            all_uris.append(f.get_uri())
        for uri in filter(
                lambda p: p.lower().endswith(self.__MIMES), all_uris):
            return uri
        infos.close(None)
        return None

    def __emit_update(self, album_id):
        """
            Announce album cover update
            @param album_id as int
        """
        if album_id is not None:
            emit_signal(self, "album-artwork-changed", album_id)

    def __update_uri(self, album):
        """
            Check if album uri exists, update if not
            @param album as Album
        """
        if not album.storage_type & StorageType.COLLECTION:
            return
        d = Gio.File.new_for_uri(album.uri)
        if not d.query_exists():
            if album.tracks:
                track_uri = album.tracks[0].uri
                f = Gio.File.new_for_uri(track_uri)
                p = f.get_parent()
                parent_uri = "" if p is None else p.get_uri()
                album.set_uri(parent_uri)

    def __save_web(self, album, data):
        """
            Save artwork for a web album
            @param album as Album
            @param data as bytes
        """
        store_path = "%s/%s" % (ALBUMS_WEB_PATH, album.lp_album_id)
        store_path = self.add_extension(store_path)
        self.save_pixbuf_from_data(store_path, data)
        self.__emit_update(album.id)

    def __save_ro(self, album, data):
        """
            Save artwork for a read only album
            @param album as Album
            @param data as bytes
        """
        store_path = "%s/%s" % (ALBUMS_PATH, album.lp_album_id)
        store_path = self.add_extension(store_path)
        self.save_pixbuf_from_data(store_path, data)
        self.__emit_update(album.id)

    def __save(self, album, data):
        """
            Save artwork for an album
            @param album as Album
            @param data as bytes
        """
        store_path = "%s/%s" % (ALBUMS_PATH, album.lp_album_id)
        store_path = self.add_extension(store_path)
        save_to_tags = App().settings.get_value("save-to-tags")
        # Multiple albums at same path
        uri_count = App().albums.get_uri_count(album.uri)
        art_uri = "%s/%s" % (album.uri, self.__favorite)
        art_uri = self.add_extension(art_uri)
        # Save cover to tags
        if save_to_tags and data is not None:
            helper = TaskHelper()
            helper.run(self.__add_to_tags, album, data)
        # We need to remove favorite if exists
        if uri_count > 1 or save_to_tags:
            f = Gio.File.new_for_uri(art_uri)
            if f.query_exists():
                f.trash()
        # Name file with album information
        if uri_count > 1:
            art_uri = "%s/%s" % (album.uri, album.lp_album_id)
            art_uri = self.add_extension(art_uri)
        self.save_pixbuf_from_data(store_path, data)
        # Keep file in store if empty
        if data is None:
            dst = Gio.File.new_for_uri(art_uri)
            if dst.query_exists():
                try:
                    dst.delete(None)
                except:
                    pass
        else:
            dst = Gio.File.new_for_uri(art_uri)
            src = Gio.File.new_for_path(store_path)
            src.move(dst, Gio.FileCopyFlags.OVERWRITE, None, None)
        self.__emit_update(album.id)

    def __get_pixbuf_from_tags(self, uri):
        """
            Return cover from tags
            @param uri as str
        """
        pixbuf = None
        # Internal URI are just like sp:
        if uri.find(":/") == -1:
            return
        try:
            discoverer = Discoverer()
            info = discoverer.get_info(uri)
            exist = False
            if info is not None:
                (exist, sample) = info.get_tags().get_sample_index("image", 0)
                if not exist:
                    (exist, sample) = info.get_tags().get_sample_index(
                        "preview-image", 0)
            if exist:
                (exist, mapflags) = sample.get_buffer().map(Gst.MapFlags.READ)
            if exist:
                bytes = GLib.Bytes.new(mapflags.data)
                stream = Gio.MemoryInputStream.new_from_bytes(bytes)
                pixbuf = GdkPixbuf.Pixbuf.new_from_stream(stream, None)
                stream.close()
        except Exception as e:
            Logger.error("AlbumArtwork::__get_pixbuf_from_tags(): %s" % e)
        return pixbuf

    def __add_to_tags(self, album, data):
        """
            Add image data to album tags
            @param album as Album
            @param data as bytes
        """
        # https://bugzilla.gnome.org/show_bug.cgi?id=747431
        bytes = GLib.Bytes.new(data)
        stream = Gio.MemoryInputStream.new_from_bytes(bytes)
        pixbuf = GdkPixbuf.Pixbuf.new_from_stream_at_scale(stream,
                                                           ArtSize.MPRIS,
                                                           ArtSize.MPRIS,
                                                           True,
                                                           None)
        stream.close()
        if self.extension == StoreExtention.PNG:
            cache_path = "%s/lollypop_cover_tags.png" % CACHE_PATH
            pixbuf.savev(cache_path, "png", [None], [None])
        else:
            cache_path = "%s/lollypop_cover_tags.jpg" % CACHE_PATH
            pixbuf.savev(cache_path, "jpeg", ["quality"], [100])
        self.__write_image_to_tags(cache_path, album)

    def __write_image_to_tags(self, cache_path, album):
        """
            Save album at path to album tags
            @param cache_path as str
            @param album as Album
        """
        files = []
        for track in album.tracks:
            App().tracks.set_mtime(track.id, int(time()) + 10)
            f = Gio.File.new_for_uri(track.uri)
            if f.query_exists():
                files.append(f.get_path())
        worked = False
        arguments = [["kid3-cli", "-c", "set picture:'%s' ''" % cache_path],
                     ["flatpak-spawn", "--host", "kid3-cli",
                      "-c", "set picture:'%s' ''" % cache_path]]
        for argv in arguments:
            argv += files
            try:
                (pid, stdin, stdout, stderr) = GLib.spawn_async(
                    argv, flags=GLib.SpawnFlags.SEARCH_PATH |
                    GLib.SpawnFlags.STDOUT_TO_DEV_NULL,
                    standard_input=False,
                    standard_output=False,
                    standard_error=False
                )
                GLib.spawn_close_pid(pid)
                worked = True
                break
            except Exception as e:
                Logger.error("AlbumArtwork::__write_image_to_tags(): %s" % e)
        if worked:
            self.clean(album)
            GLib.timeout_add(2000, self.__emit_update, album.id)
        else:
            App().notify.send("Lollypop",
                              _("You need to install kid3-cli"))



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, GObject

from time import time
import json

from lollypop.logger import Logger
from lollypop.utils import emit_signal
from lollypop.utils import get_lollypop_album_id, get_lollypop_track_id
from lollypop.objects_album import Album
from lollypop.define import App, Type
from lollypop.collection_item import CollectionItem


class SaveWebHelper(GObject.Object):
    """
       Web helper for saving Spotify payloads
    """

    __gsignals__ = {
        "match-album": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-track": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-artist": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        """
            Init helper
        """
        GObject.Object.__init__(self)

    def save_track_payload_to_db(self, payload, item,
                                 storage_type, notify, cancellable):
        """
            Save track to DB
            @param payload as {}
            @param item as CollectionItem
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
            @param notify as bool
        """
        lp_track_id = get_lollypop_track_id(payload["name"],
                                            payload["artists"],
                                            item.album_name)
        item.track_id = App().tracks.get_id_for_lp_track_id(lp_track_id)
        if item.track_id < 0:
            self.__save_track(payload, item, storage_type)
        if notify:
            emit_signal(self, "match-track", item.track_id, storage_type)

    def save_album_payload_to_db(self, payload, storage_type,
                                 notify, cancellable):
        """
            Save album to DB
            @param payload as {}
            @param storage_type as StorageType
            @param notify as bool
            @param cancellable as Gio.Cancellable
            @return CollectionItem/None
        """
        lp_album_id = get_lollypop_album_id(payload["name"],
                                            payload["artists"])
        album_id = App().albums.get_id_for_lp_album_id(lp_album_id)
        if album_id >= 0:
            album = Album(album_id)
            if notify:
                emit_signal(self, "match-album", album_id, storage_type)
            return album.collection_item
        item = self.__save_album(payload, storage_type)
        album = Album(item.album_id)
        if notify:
            App().album_art.add_from_uri(album,
                                         payload["artwork-uri"],
                                         cancellable)
            emit_signal(self, "match-album", album.id, storage_type)
        return item

#######################
# PRIVATE             #
#######################
    def __get_date_from_payload(self, payload):
        """
            Get date from payload
            @param payload as {}
            @param return (int, int)
        """
        try:
            release_date = payload["date"]
            dt = GLib.DateTime.new_from_iso8601(release_date,
                                                GLib.TimeZone.new_local())
            return (dt.to_unix(), dt.get_year())
        except:
            pass
        return (None, None)

    def __get_cover_art_uri(self, mbid, cancellable):
        """
            Get cover art URI for mbid
            @param mbid as str
            @param cancellable as Gio.Cancellable
            @return str/None
        """
        try:
            uri = "http://coverartarchive.org/release/%s" % mbid
            (status, data) = App().task_helper.load_uri_content_sync(uri, None)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for image in decode["images"]:
                    if not image["front"]:
                        continue
                    return image["image"]
        except Exception as e:
            Logger.error(e)
            Logger.error(
                "SaveWebHelper::__get_cover_art_uri(): %s", data)
        return None

    def __save_album(self, payload, storage_type):
        """
            Save album payload to DB
            @param payload as {}
            @param storage_type as StorageType
            @return CollectionItem
        """
        (timestamp, year) = self.__get_date_from_payload(payload)
        item = CollectionItem(uri=payload["uri"],
                              album_artists=payload["artists"],
                              album_name=payload["name"],
                              album_mtime=int(time()),
                              year=year,
                              timestamp=timestamp,
                              mb_album_id=payload["mbid"],
                              album_synced=payload["track-count"],
                              storage_type=storage_type)
        Logger.debug("SaveWebHelper::save_album(): %s - %s",
                     item.album_artists, item.album_name)
        App().scanner.save_album(item)
        App().albums.add_genre(item.album_id, Type.WEB)
        return item

    def __save_track(self, payload, item, storage_type):
        """
            Save track payload to DB
            @param payload as {}
            @param storage_type as StorageType
            @return track_id as int
        """
        item.track_name = payload["name"]
        item.artists = payload["artists"]
        Logger.debug("SaveWebHelper::save_track(): %s - %s",
                     item.artists, item.track_name)
        item.discnumber = int(payload["discnumber"])
        item.tracknumber = int(payload["tracknumber"])
        item.duration = payload["duration"]
        item.track_mtime = int(time())
        item.uri = payload["uri"]
        item.mb_track_id = payload["mbid"]
        item.storage_type = storage_type
        App().scanner.save_track(item)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _
from hashlib import sha256

from lollypop.define import App
from lollypop.objects_track import Track
from lollypop.objects_album import Album


class PlaylistsMenu(Gio.Menu):
    """
        Contextual menu for playlists
    """

    def __init__(self, objects):
        """
            Init playlist menu
            @param objects as [Track/Album]
        """
        Gio.Menu.__init__(self)
        self.__objects = objects
        self.__set_playlist_actions()

#######################
# PRIVATE             #
#######################
    def __set_playlist_actions(self):
        """
            Set playlist actions
        """
        add_action = Gio.SimpleAction(name="add_to_new")
        App().add_action(add_action)
        add_action.connect("activate", self.__on_add_action_activate)
        self.append(_("New playlist"), "app.add_to_new")
        i = 1
        exists = True
        for (playlist_id, name) in App().playlists.get():
            if App().playlists.get_smart(playlist_id):
                continue
            for obj in self.__objects:
                if isinstance(obj, Album):
                    exists = App().playlists.exists_album(playlist_id, obj)
                else:
                    exists = App().playlists.exists_track(playlist_id, obj.uri)
                if not exists:
                    break
            encoded = sha256(name.encode("utf-8")).hexdigest()
            playlist_action = Gio.SimpleAction.new_stateful(
                                          encoded,
                                          None,
                                          GLib.Variant.new_boolean(exists))
            playlist_action.connect("change-state",
                                    self.__on_playlist_action_change_state,
                                    playlist_id)
            App().add_action(playlist_action)
            self.append(name, "app.%s" % encoded)
            i += 1

    def __add_to_playlist(self, playlist_id):
        """
            Add to playlist
            @param playlist_id as int
        """
        def add(playlist_id):
            tracks = []
            for obj in self.__objects:
                if isinstance(obj, Album):
                    for track_id in obj.track_ids:
                        tracks.append(Track(track_id))
                else:
                    tracks = [Track(obj.id)]
                App().playlists.add_tracks(playlist_id, tracks, True)
        App().task_helper.run(add, playlist_id)

    def __remove_from_playlist(self, playlist_id):
        """
            Del from playlist
            @param playlist_id as int
        """
        def remove(playlist_id):
            tracks = []
            for obj in self.__objects:
                if isinstance(obj, Album):
                    for track_id in obj.track_ids:
                        tracks.append(Track(track_id))
                else:
                    tracks = [Track(obj.id)]
                App().playlists.remove_tracks(playlist_id, tracks, True)
        App().task_helper.run(remove, playlist_id)

    def __on_playlist_action_change_state(self, action, variant, playlist_id):
        """
            Add/Remove object from playlist
            @param Gio.SimpleAction
            @param GLib.Variant
            @param playlist_id as int
        """
        action.set_state(variant)
        if variant:
            self.__add_to_playlist(playlist_id)
        else:
            self.__remove_from_playlist(playlist_id)

    def __on_add_action_activate(self, action, variant):
        """
            Add a new playlist and add object to playlist
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        playlist_id = App().playlists.add(App().playlists.get_new_name())
        self.__add_to_playlist(playlist_id)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# This is global object initialised at lollypop start
# member init order is important!

from gi.repository import Gio, GLib


App = Gio.Application.get_default

LASTFM_API_KEY = "7a9619a850ccf7377c46cf233c51e3c6"
LASTFM_API_SECRET = "9254319364d73bec6c59ace485a95c98"

DISCOGS_API_KEY = "vcKgyONtKsaQZLvpUfOU"
DISCOGS_API_SECRET = "SmWphfGWALsmDnoyhPuQlDDQEIvFJuPM"

GOOGLE_API_ID = "015987506728554693370:waw3yqru59a"

MARGIN_BIG = 30
MARGIN = 15
MARGIN_MEDIUM = 10
MARGIN_SMALL = 5

LOLLYPOP_DATA_PATH = GLib.get_user_data_dir() + "/lollypop"
# All cache goes here
CACHE_PATH = GLib.get_user_cache_dir() + "/lollypop"
# Stores for albums
ALBUMS_PATH = LOLLYPOP_DATA_PATH + "/albums"
ALBUMS_WEB_PATH = LOLLYPOP_DATA_PATH + "/albums_web"
# Stores for artists
ARTISTS_PATH = LOLLYPOP_DATA_PATH + "/artists"
# Store for lyrics
LYRICS_PATH = LOLLYPOP_DATA_PATH + "/lyrics"


class TimeStamp:
    ONE_YEAR = 31536000
    TWO_YEAR = 63072000
    THREE_YEAR = 94608000


class FileType:
    UNKNOWN = 0
    AUDIO = 1
    PLS = 2
    OTHER = 3


class Repeat:
    NONE = 0
    AUTO_SIMILAR = 1
    AUTO_RANDOM = 2
    TRACK = 3
    ALL = 4


class LovedFlags:
    NONE = 1 << 0
    LOVED = 1 << 1
    SKIPPED = 1 << 2


class GstPlayFlags:
    GST_PLAY_FLAG_VIDEO = 1 << 0  # We want video output
    GST_PLAY_FLAG_AUDIO = 1 << 1  # We want audio output
    GST_PLAY_FLAG_TEXT = 1 << 3   # We want subtitle output


class StorageType:
    NONE = 1 << 0
    COLLECTION = 1 << 1
    EPHEMERAL = 1 << 2
    SAVED = 1 << 3
    SPOTIFY_NEW_RELEASES = 1 << 4
    SPOTIFY_SIMILARS = 1 << 5
    EXTERNAL = 1 << 6
    SEARCH = 1 << 7
    DEEZER_CHARTS = 1 << 8
    ALL = 1 << 1 | 1 << 2 | 1 << 3 | 1 << 4 | 1 << 5 | 1 << 6 | 1 << 7 | 1 << 8


class ArtBehaviour:
    NONE = 1 << 0
    ROUNDED = 1 << 1
    ROUNDED_BORDER = 1 << 2
    BLUR = 1 << 3
    BLUR_HARD = 1 << 4
    BLUR_MAX = 1 << 5
    FALLBACK = 1 << 6
    DARKER = 1 << 7
    LIGHTER = 1 << 8
    CROP = 1 << 9
    CROP_SQUARE = 1 << 10
    CACHE = 1 << 11
    NO_CACHE = 1 << 12


class ViewType:
    DEFAULT = 1 << 0
    TWO_COLUMNS = 1 << 1
    SINGLE_COLUMN = 1 << 2
    DND = 1 << 3
    SEARCH = 1 << 4
    PLAYLISTS = 1 << 5
    ALBUM = 1 << 6
    SMALL = 1 << 7
    QUEUE = 1 << 8
    SCROLLED = 1 << 9
    OVERLAY = 1 << 10
    FULLSCREEN = 1 << 11
    PLAYBACK = 1 << 12
    BANNER = 1 << 13
    ARTIST = 1 << 14
    TOOLBAR = 1 << 15


NetworkAccessACL = {
    "DATA": 1 << 1,
    "LASTFM": 1 << 2,
    "SPOTIFY": 1 << 3,
    "YOUTUBE": 1 << 4,
    "GOOGLE": 1 << 5,
    "STARTPAGE": 1 << 6,
    "WIKIPEDIA": 1 << 7,
    "JAMENDO": 1 << 8,
    "MUSICBRAINZ": 1 << 9,
    "ITUNES": 1 << 10,
    "DEEZER": 1 << 11,
    "WIKIA": 1 << 12,
    "GENIUS": 1 << 13,
    "AUDIODB": 1 << 14,
    "FANARTTV": 1 << 15,
    "DUCKDUCKGO": 1 << 16,
    "LIBREFM": 1 << 17,
    "METROLYRICS": 1 << 18,
    "BING": 1 << 19,
}


class StoreExtention:
    JPG = 0
    PNG = 1


class LoadingState:
    NONE = 0
    RUNNING = 1
    ABORTED = 2
    FINISHED = 3


class IndicatorType:
    NONE = 1 << 0
    PLAY = 1 << 1
    LOVED = 1 << 2
    SKIPPED = 1 << 3
    LOADING = 1 << 4


class ArtSize:
    SMALL = 50     # Depends on cover-size in gsettings
    MEDIUM = 100
    BANNER = 150   # Depends on cover-size in gsettings
    BIG = 200      # Depends on cover-size in gsettings
    MINIPLAYER = 300
    FULLSCREEN = 400
    MPRIS = 900


class ScanType:
    EXTERNAL = 0
    NEW_FILES = 1
    FULL = 2


class ScanUpdate:
    ADDED = 0
    REMOVED = 1
    MODIFIED = 2


class SelectionListMask:
    NONE = 1 << 0
    SIDEBAR = 1 << 1
    FASTSCROLL = 1 << 2
    ARTISTS = 1 << 3
    GENRES = 1 << 4
    PLAYLISTS = 1 << 5
    COMPILATIONS = 1 << 6
    LABEL = 1 << 7
    ELLIPSIZE = 1 << 8


class Notifications:
    NONE = 0
    ALL = 1
    MPRIS = 2


class PowerManagement:
    NONE = 0             # Use OS defaults
    IDLE = 1             # Inhibit screensaver
    SUSPEND = 2          # Inhibit suspend
    BOTH = 3             # Inhibit screensaver and suspend


class ReplayGain:
    NONE = 0
    TRACK = 1
    ALBUM = 1


class Size:
    MINI = 250
    PHONE = 360  # Librem Phone
    SMALL = 500
    MEDIUM = 720
    NORMAL = 850
    BIG = 1000


class OrderBy:
    ARTIST_YEAR = 0
    ARTIST_TITLE = 1
    TITLE = 2
    YEAR_DESC = 3
    POPULARITY = 4
    YEAR_ASC = 5


# Order is important
class Type:
    NONE = -1
    SUGGESTIONS = -2
    POPULARS = -3
    RANDOMS = -4
    RECENTS = -5
    LOVED = -6
    LITTLE = -7
    SKIPPED = -8
    # WEB is stored in DB, can't be changed
    WEB = -9
    # Stored in DB, can't be changed
    COMPILATIONS = -10
    ARTISTS = -11
    ARTISTS_LIST = -12
    GENRES = -13
    GENRES_LIST = -14
    YEARS = -15
    PLAYLISTS = -16
    SMART = -19
    EQUALIZER = -20
    DEVICE_ALBUMS = -27
    DEVICE_PLAYLISTS = -28
    ALBUM = -29
    ALL = -99
    SEPARATOR = -100
    CURRENT = -101
    INFO = -102
    SEARCH = -103
    LYRICS = -104


LATIN1_ENCODING = b"\x00"
"""Byte code for latin1"""
UTF_16_ENCODING = b"\x01"
"""Byte code for UTF-16"""
UTF_16BE_ENCODING = b"\x02"
"""Byte code for UTF-16 (big endian)"""
UTF_8_ENCODING = b"\x03"
"""Byte code for UTF-8 (Not supported in ID3 versions < 2.4)"""


SPOTIFY_CLIENT_ID = "0b144843878a46b2b12e0958c342c3ac"
SPOTIFY_SECRET = "265ab8e057684f1b9e69e0c58f4881c1"
AUDIODB_CLIENT_ID = "195003"
FANARTTV_ID = "1cb10bc910c8d4fc34e0c78ac4e8ef46"
